import cv2
import numpy as np

import zebrazoom.code.util as util


folderName = 'allCatamaran'

cap1 = zzVideoReading.VideoCapture(folderName + '/cluster1.avi')
cap2 = zzVideoReading.VideoCapture(folderName + '/cluster3.avi')
cap3 = zzVideoReading.VideoCapture(folderName + '/cluster4.avi')
cap4 = zzVideoReading.VideoCapture(folderName + '/cluster5.avi')

frame_width  = int(cap1.get(3))
frame_height = int(cap1.get(4))
max1         = int(cap1.get(7))
max2         = int(cap2.get(7))
max3         = int(cap3.get(7))
max4         = int(cap4.get(7))
minOfMaxs    = min(max1, min(max2, min(max3, max4)))

out = cv2.VideoWriter('outpy.avi',cv2.VideoWriter_fourcc('M','J','P','G'), 10, (2*frame_width, 2*frame_height))
# out = cv2.VideoWriter('outpy.avi',cv2.VideoWriter_fourcc('M','J','P','G'), 10, (frame_width, frame_height))

affiche = False

i = 0
# Read until video is completed
while(cap1.isOpened() and (i < minOfMaxs - 1)):
  i = i + 1
  # Capture frame-by-frame
  ret, frame1 = cap1.read()
  ret, frame2 = cap2.read()
  ret, frame3 = cap3.read()
  ret, frame4 = cap4.read()
  
  if ret == True:
    # frameF = np.zeros((2*frame_width, 2*frame_height, 3))
    # frameF[0:frame_height,              0:frame_width]             = frame1
    # frameF[0:frame_height,              frame_width:2*frame_width] = frame2
    # frameF[frame_height:2*frame_height, 0:frame_width]             = frame3
    # frameF[frame_height:2*frame_height, frame_width:2*frame_width] = frame4

    # frameF = np.zeros((frame_width, frame_height, 3))
    # frameF[0:frame_height, 0:frame_width] = frame1
    fl1    = np.concatenate((frame1, frame2), axis=1)
    fl2    = np.concatenate((frame3, frame4), axis=1)
    frameF = np.concatenate((fl1, fl2), axis=0)
    
    font = cv2.FONT_HERSHEY_SIMPLEX
    ydown = 13
    fontSize = 0.4
    cv2.putText(frameF,"Cluster 1:",(int(1),int(10)),font,fontSize,(255,255,255))
    cv2.putText(frameF,"Slow Forward Swims",(int(1),int(10+ydown)),font,fontSize,(255,255,255))
    
    cv2.putText(frameF,"Cluster 2:",(int(2+frame_width),int(10)),font,fontSize,(255,255,255))
    cv2.putText(frameF,"Small Amplitude Turns",(int(2+frame_width),int(10+ydown)),font,fontSize,(255,255,255))
    
    cv2.putText(frameF,"Cluster 3:",(int(1),int(frame_height+11)),font,fontSize,(255,255,255))
    cv2.putText(frameF,"Large Amplitude Turns",(int(1),int(frame_height+11+ydown)),font,fontSize,(255,255,255))
    
    cv2.putText(frameF,"Cluster 4:",(int(2+frame_width),int(frame_height+11)),font,fontSize,(255,255,255))
    cv2.putText(frameF,"Burst Swims",(int(2+frame_width),int(frame_height+11+ydown)),font,fontSize,(255,255,255))
    
    lineThickness = 1
    cv2.line(frameF, (frame_width, 0),  (frame_width, 2*frame_height), (255,255,255), lineThickness)
    cv2.line(frameF, (0, frame_height), (2*frame_width, frame_height), (255,255,255), lineThickness)
    
    if (affiche):
      # Display the resulting frame
      util.showFrame(frameF, title='Frame')
    else:
       out.write(frameF)
 
  # Break the loop
  else: 
    break

# When everything done, release the video capture object
cap1.release()
cap2.release()
cap3.release()
cap4.release()
 
# Closes all the frames
cv2.destroyAllWindows()



import scipy.io
import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import matplotlib.pyplot as plt
from sklearn.manifold import TSNE
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture
from zebrazoom.dataAnalysis.dataanalysis.visualizeClusters import visualizeClusters
from zebrazoom.dataAnalysis.dataanalysis.outputValidationVideo import outputValidationVideo
from zebrazoom.dataAnalysis.dataanalysis.activeLearning import prepareForActiveLearning
from scipy.stats import chi2
import cv2
import os
import shutil
import pickle

def applyClustering(clusteringOptions, classifier, outputFolder, ZZoutputLocation=''):

  pca = 0
  kme = 0
  if classifier:
    print("reloading classifier")
    pca = classifier[0]
    model  = classifier[1]

  analyzeAllWellsAtTheSameTime   = clusteringOptions['analyzeAllWellsAtTheSameTime']
  pathToVideos                   = clusteringOptions['pathToVideos']
  nbCluster                      = clusteringOptions['nbCluster']
  if 'nbPcaComponents' in clusteringOptions:
    nbPcaComponents              = clusteringOptions['nbPcaComponents']
  else:
    nbPcaComponents              = 0
  nbFramesTakenIntoAccount       = clusteringOptions['nbFramesTakenIntoAccount']
  scaleGraphs                    = clusteringOptions['scaleGraphs']
  showFigures                    = clusteringOptions['showFigures']
  useFreqAmpAsym                 = clusteringOptions['useFreqAmpAsym']
  useAngles                      = clusteringOptions['useAngles']
  useAnglesSpeedHeadingDisp      = clusteringOptions['useAnglesSpeedHeadingDisp']
  useAnglesSpeedHeading          = clusteringOptions['useAnglesSpeedHeading']
  useAnglesSpeed                 = clusteringOptions['useAnglesSpeed']
  useAnglesHeading               = clusteringOptions['useAnglesHeading']
  useAnglesHeadingDisp           = clusteringOptions['useAnglesHeadingDisp']
  useFreqAmpAsymSpeedHeadingDisp = clusteringOptions['useFreqAmpAsymSpeedHeadingDisp']
  useAngleAnd3GlobalParameters   = clusteringOptions['useAngleAnd3GlobalParameters'] if 'useAngleAnd3GlobalParameters' in clusteringOptions else 0
  videoSaveFirstTenBouts         = clusteringOptions['videoSaveFirstTenBouts']
  nbVideosToSave                 = clusteringOptions['nbVideosToSave']
  resFolder                      = clusteringOptions['resFolder']
  nameOfFile                     = clusteringOptions['nameOfFile']
  globalParametersCalculations   = clusteringOptions['globalParametersCalculations']
  
  modelUsedForClustering = clusteringOptions['modelUsedForClustering'] if 'modelUsedForClustering' in clusteringOptions else 'KMeans'
  
  removeOutliers = clusteringOptions['removeOutliers'] if 'removeOutliers' in clusteringOptions else False
  
  removeBoutsContainingNanValuesInParametersUsedForClustering = clusteringOptions['removeBoutsContainingNanValuesInParametersUsedForClustering'] if 'removeBoutsContainingNanValuesInParametersUsedForClustering' in clusteringOptions else True
  
  instaTBF   = ['instaTBF'+str(i)  for i in range(1,nbFramesTakenIntoAccount+1)]
  instaAmp   = ['instaAmp'+str(i)  for i in range(1,nbFramesTakenIntoAccount+1)]
  instaAsym  = ['instaAsym'+str(i) for i in range(1,nbFramesTakenIntoAccount+1)]

  tailAngles = ['tailAngles'+str(i) for i in range(1,nbFramesTakenIntoAccount+1)]

  instaSpeed       = ['instaSpeed'       + str(i) for i in range(1,nbFramesTakenIntoAccount+1)]
  instaHeadingDiff = ['instaHeadingDiff' + str(i) for i in range(1,nbFramesTakenIntoAccount+1)]
  instaHorizDispl  = ['instaHorizDispl'  + str(i) for i in range(1,nbFramesTakenIntoAccount+1)]
  
  allInstas  = instaTBF + instaAmp + instaAsym

  allInstas2 = tailAngles + instaSpeed + instaHeadingDiff + instaHorizDispl

  xaxis    = [0, 30]
  freqAxis = [0, 0.5]
  ampAxis  = [0, 1.6]
  asymAxis = [0, 0.8]
  angAxis  = [0, 1.5]

  possibleColors = ['b', 'r', 'g', 'k', 'c', 'm', 'y']
  possibleColorsNames = ['blue', 'red', 'green', 'black', 'cyan', 'magenta', 'yellow']
  
  outputFolderResult = os.path.join(outputFolder, nameOfFile)
  
  if os.path.exists(outputFolderResult):
    shutil.rmtree(outputFolderResult)
  os.mkdir(outputFolderResult)
  
  infile = open(os.path.join(resFolder, nameOfFile + '.pkl'),'rb')
  dfParam = pickle.load(infile)
  infile.close()
  
  nbConditions = len(np.unique(dfParam['Condition'].values))

  # Applying PCA
  if classifier == 0:
    print("creating pca object")
    if nbPcaComponents:
      pca = PCA(n_components = nbPcaComponents)
    else:
      pca = PCA()

  if useFreqAmpAsym:
    allInstaValues = dfParam[allInstas].values

  if useAngles:
    allInstaValues = dfParam[tailAngles].values

  if useAnglesSpeedHeadingDisp:
    allInstaValues = dfParam[tailAngles + instaSpeed + instaHeadingDiff + instaHorizDispl].values
    
  if useAnglesSpeedHeading:
    allInstaValues = dfParam[tailAngles + instaSpeed + instaHeadingDiff].values
    
  if useAnglesSpeed:
    allInstaValues = dfParam[tailAngles + instaSpeed].values

  if useAnglesHeading:
    allInstaValues = dfParam[tailAngles + instaHeadingDiff].values
    
  if useAnglesHeadingDisp:
    allInstaValues = dfParam[tailAngles + instaHeadingDiff + instaHorizDispl].values
    
  if useFreqAmpAsymSpeedHeadingDisp:
    allInstaValues = dfParam[allInstas + instaSpeed + instaHeadingDiff + instaHorizDispl].values
  
  if useAngleAnd3GlobalParameters:
    allInstaValues = dfParam[tailAngles].values
  
  # if modelUsedForClustering == 'KMeans':
  scaler = StandardScaler()
  allInstaValues = scaler.fit_transform(allInstaValues)
  
  if removeBoutsContainingNanValuesInParametersUsedForClustering:
    allInstaValuesLenBef = len(allInstaValues)
    try:
      dfParam = dfParam.drop([idx for idx, val in enumerate(~np.isnan(allInstaValues).any(axis=1)) if not(val)])
    except:
      scaler = StandardScaler()
      allInstaValues = scaler.fit_transform(allInstaValues)
      dfParam = dfParam.drop([idx for idx, val in enumerate(~np.isnan(allInstaValues).any(axis=1)) if not(val)])
    allInstaValues = allInstaValues[~np.isnan(allInstaValues).any(axis=1)]
    allInstaValuesLenAft = len(allInstaValues)
    if allInstaValuesLenBef - allInstaValuesLenAft > 0:
      print(allInstaValuesLenBef - allInstaValuesLenAft, " bouts (out of ", allInstaValuesLenBef, " ) were deleted because they contained NaN values")
    else:
      print("all bouts were kept (no nan values)")
  else:
    dfParam = dfParam.fillna(0)
    allInstaValues = np.nan_to_num(allInstaValues)
    print("nan values replaced by zeros")
  
  if 'level_0' in dfParam.columns:
    dfParam = dfParam.drop(['level_0'], axis=1)
  
  dfParam = dfParam.reset_index()
  if removeOutliers:
    covariance  = np.cov(allInstaValues , rowvar=False)
    covariance_pm1 = np.linalg.matrix_power(covariance, -1)
    centerpoint = np.mean(allInstaValues , axis=0)
    distances = []
    for i, val in enumerate(allInstaValues):
      p1 = val
      p2 = centerpoint
      distance = (p1-p2).T.dot(covariance_pm1).dot(p1-p2)
      distances.append(distance)
    distances = np.array(distances)
    cutoff = chi2.ppf(0.95, allInstaValues.shape[1])
    outlierIndexes = np.where(distances < cutoff )
    nbBoutsBefore = len(dfParam)
    print("Number of bouts before outliers removal:", len(dfParam))
    dfParam = dfParam.drop([idx for idx, val in enumerate(distances >= cutoff) if val])
    nbBoutsAfter = len(dfParam)
    print("Number of bouts after outliers removal:", len(dfParam))
    print("Percentage of bouts removed:", ((nbBoutsBefore-nbBoutsAfter)/nbBoutsBefore)*100, "%")
    allInstaValues = allInstaValues[ distances < cutoff , :]
  if classifier == 0:
    print("creating pca transform and applying it on the data")
    pca_result = pca.fit_transform(allInstaValues)
  else:
    print("applying pca (reloaded)")
    pca_result = pca.transform(allInstaValues)
    
  dfParam = dfParam.drop(['level_0'], axis=1)
  dfParam = dfParam.reset_index()
  
  if useAngleAnd3GlobalParameters:
    pca_result = pca_result[:, 0:3]
    pca_result = np.concatenate((pca_result, dfParam[['deltaHead', 'Speed', 'tailAngleIntegral']].values), axis=1)
    scaler     = StandardScaler()
    pca_result = scaler.fit_transform(pca_result)  

  ind = []
  for i in range(0,nbConditions):
    ind.append(dfParam.loc[(dfParam['Condition'] == i)].index.values)
    
  # KMean clustering
  if classifier == 0:
    if modelUsedForClustering == 'KMeans':
      model = KMeans(n_clusters = nbCluster)
    elif modelUsedForClustering == 'GaussianMixture':
      model = GaussianMixture(n_components = nbCluster)
    else:
      model = KMeans(n_clusters = nbCluster)
    model.fit(pca_result)
    
  labels = model.predict(pca_result)
  if modelUsedForClustering == 'GaussianMixture':
    predictedProbas = model.predict_proba(pca_result)

  # Sorting labels
  nbLabels       = clusteringOptions['nbCluster']
  labels2        = np.zeros(len(labels))
  nbElemPerClass = np.zeros(nbLabels) 
  for i in range(0, nbLabels):
    nbElemPerClass[i] = labels.tolist().count(i)
  sortedIndices = (-nbElemPerClass).argsort()
  for i in range(0, len(labels)):
    labels2[i] = np.where(sortedIndices==labels[i])[0][0]
  dfParam['classification'] = labels2
  
  if modelUsedForClustering == 'GaussianMixture':
    for j in range(0, nbLabels):
      probasClassJ = predictedProbas[:, sortedIndices[j]]
      dfParam['classProba' + str(j)] = probasClassJ
  
  [proportions, sortedRepresentativeBouts, sortedRepresentativeBoutsIndex] = visualizeClusters(dfParam, labels2, [], modelUsedForClustering, nbConditions, nbCluster, nbFramesTakenIntoAccount, scaleGraphs, showFigures, outputFolderResult, 0, 1)
  
  if False:
    prepareForActiveLearning(proportions, sortedRepresentativeBouts, outputFolderResult, nbCluster, pca_result, dfParam, sortedRepresentativeBoutsIndex, tailAngles)
  else:
    outputFolderResult2 = os.path.join(outputFolderResult, 'savedRawData')
    if os.path.exists(outputFolderResult2):
      shutil.rmtree(outputFolderResult2)
    os.mkdir(outputFolderResult2)
    pickle.dump({'pca_result': pca_result.tolist(), 'dfParam': dfParam}, open(os.path.join(outputFolderResult2, 'boutParameters.pkl'), 'wb'))
  
  return [dfParam, [pca, model]]



import cv2
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
from zebrazoom.code.extractParameters import calculateAngle
from zebrazoom.code.extractParameters import calculateTailAngle
import h5py
import os
import shutil
import math
import matplotlib.pyplot as plt
import numpy as np
import time
import pandas as pd
import pickle
from scipy import ndimage


def calculateCurvature(TailX_VideoReferential, TailY_VideoReferential, hyperparameters):
  curvature = []
  for l in range(0, len(TailX_VideoReferential)):
    tailX = TailX_VideoReferential[l]
    tailY = TailY_VideoReferential[l]

    ydiff  = np.diff(tailY)
    ydiff2 = np.diff(ydiff)
    xdiff  = np.diff(tailX)
    xdiff2 = np.diff(xdiff)
    curv = xdiff2
    l = len(curv)
    av = 0
    for ii in range(0, l):#-1):
      num = xdiff[ii] * ydiff2[ii] - ydiff[ii] * xdiff2[ii]
      den = (xdiff[ii]**2 + ydiff[ii]**2)**1.5
      curv[ii] = num / den

    curv = curv[hyperparameters["nbPointsToIgnoreAtCurvatureBeginning"]: l-hyperparameters["nbPointsToIgnoreAtCurvatureEnd"]]
    curvature.append(curv)
  return curvature


def perBoutOutput(superStruct, hyperparameters, videoNameWithTimestamp):
  
  # Creation of the sub-folder "perBoutOutput"
  if hyperparameters["saveAllDataEvenIfNotInBouts"] or hyperparameters["saveCurvaturePlots"] or hyperparameters["saveTailAngleGraph"] or hyperparameters["saveSubVideo"] or hyperparameters["saveCurvatureData"]:
    outputPath = os.path.join(hyperparameters["outputFolder"], videoNameWithTimestamp, "perBoutOutput")
    if not os.path.exists(outputPath):
      os.makedirs(outputPath)
      
  # opening the video previously created in createValidationVideo as an input video stream
  if hyperparameters['storeH5']:
    videoPath = os.path.join(hyperparameters["outputFolder"], f'{videoNameWithTimestamp}.avi')
  else:
    videoPath = os.path.join(hyperparameters["outputFolder"], hyperparameters["videoName"], f'{videoNameWithTimestamp}.avi')
  cap = zzVideoReading.VideoCapture(videoPath)
  
  if (cap.isOpened()== False): 
    print("Error opening video stream or file")
  
  # Video frame start and end
  firstFrame = hyperparameters["firstFrame"]
  lastFrame  = hyperparameters["lastFrame"]
  nbFrames   = lastFrame - firstFrame + 1
  perBoutOutputVideoStartStopFrameMargin = hyperparameters["perBoutOutputVideoStartStopFrameMargin"]
  
  # Getting well positions
  infoWells = []
  for i in range(0, len(superStruct["wellPositions"])):
    x = superStruct["wellPositions"][i]["topLeftX"]
    y = superStruct["wellPositions"][i]["topLeftY"]
    lengthX = superStruct["wellPositions"][i]["lengthX"]
    lengthY = superStruct["wellPositions"][i]["lengthY"]
    infoWells.append([x, y, lengthX, lengthY])
  
  # alternativeCurvatureCalculation: Going through each well, each fish and each bout
  if hyperparameters["alternativeCurvatureCalculation"]:
    for i in range(0, len(superStruct["wellPoissMouv"])):
      for j in range(0, len(superStruct["wellPoissMouv"][i])):
        for k in range(0, len(superStruct["wellPoissMouv"][i][j])):
          
          # Creation of the curvature graph for bout k
          TailX_VideoReferential = superStruct["wellPoissMouv"][i][j][k]["TailX_VideoReferential"]
          TailY_VideoReferential = superStruct["wellPoissMouv"][i][j][k]["TailY_VideoReferential"]
          
          alternativeCurvatureCalculation = []
          for l in range(0, len(TailX_VideoReferential)):
            tailX = TailX_VideoReferential[l]
            tailY = TailY_VideoReferential[l]

            l = len(tailX)
            alternativeCurvatureCalculationForFrame = np.zeros(l-2)
            av = 0
            for ii in range(1, l-1):
              angleBef = calculateAngle(np.array([tailX[ii-1], tailY[ii-1]]), np.array([tailX[ii],   tailY[ii]]))
              angleAft = calculateAngle(np.array([tailX[ii],   tailY[ii]]),   np.array([tailX[ii+1], tailY[ii+1]]))
              alternativeCurvatureCalculationForFrame[ii-1] = calculateTailAngle(angleBef, angleAft)
            alternativeCurvatureCalculation.append(alternativeCurvatureCalculationForFrame)
          
          rolling_window = hyperparameters["curvatureMedianFilterSmoothingWindow"]
          if rolling_window:
            alternativeCurvatureCalculation = ndimage.median_filter(alternativeCurvatureCalculation, size=rolling_window) # 2d median filter
          else:
            alternativeCurvatureCalculation = np.array(alternativeCurvatureCalculation)
          
          alternativeCurvatureCalculation = np.flip(np.transpose(alternativeCurvatureCalculation), 0)
          
          superStruct["wellPoissMouv"][i][j][k]["alternativeCurvatureCalculation"] = alternativeCurvatureCalculation.tolist()
          
          if hyperparameters["saveCurvaturePlots"]:
            fig = plt.figure(1)
            if hyperparameters["maxCurvatureValues"] == 0:
              maxx = max([max(abs(np.array(t))) for t in alternativeCurvatureCalculation])
            else:
              maxx = hyperparameters["maxCurvatureValues"]
            if hyperparameters["curvatureXaxisNbFrames"] != 0:
              if len(alternativeCurvatureCalculation[0]) < hyperparameters["curvatureXaxisNbFrames"]:
                alternativeCurvatureCalculation2 = np.pad(alternativeCurvatureCalculation, ((0, 0), (0, hyperparameters["curvatureXaxisNbFrames"] - len(alternativeCurvatureCalculation[0]))), mode='constant')
              elif len(alternativeCurvatureCalculation[0]) > hyperparameters["curvatureXaxisNbFrames"]:
                alternativeCurvatureCalculation2 = alternativeCurvatureCalculation[:, 0:hyperparameters["curvatureXaxisNbFrames"]]
              else:
                alternativeCurvatureCalculation2 = alternativeCurvatureCalculation
            else:
              alternativeCurvatureCalculation2 = alternativeCurvatureCalculation
            plt.pcolor(alternativeCurvatureCalculation2, vmin=-maxx, vmax=maxx, cmap=hyperparameters["colorMapCurvature"])
            ax = fig.axes
            ax[0].set_ylabel('Rostral to Caudal')
            if hyperparameters["videoFPS"]:
              ax[0].set_xlabel('Second')
              plt.xticks([i for i in range(0, len(alternativeCurvatureCalculation2[0]), int(len(alternativeCurvatureCalculation2[0])/10))], [int(100*(i/hyperparameters["videoFPS"]))/100 for i in range(0, len(alternativeCurvatureCalculation2[0]), int(len(alternativeCurvatureCalculation2[0])/10))])
            else:
              ax[0].set_xlabel('Frame number')
            
            if hyperparameters["videoPixelSize"] and int(len(alternativeCurvatureCalculation2)/10):
              ax[0].set_ylabel('Rostral to Caudal (in mm)')
              plt.yticks([i for i in range(0, len(alternativeCurvatureCalculation2), int(len(alternativeCurvatureCalculation2)/10))], [int(100*( hyperparameters["videoPixelSize"] * tailLenghtInPixels * (i/len(alternativeCurvatureCalculation2)) ))/100 for i in range(0, len(alternativeCurvatureCalculation2), int(len(alternativeCurvatureCalculation2)/10))])
            else:
              ax[0].set_ylabel('Rostral to Caudal (arbitrary units)')
            
            plt.colorbar()
            plt.savefig(os.path.join(outputPath, hyperparameters["videoName"] + "_alternativeCurvatureCalculation_bout" + str(i) + '_' + str(j) + '_' + str(k) + '.png'))
            plt.close(1)
  
  
  # Curvature calculation: Going through each well, each fish and each bout
  for i in range(0, len(superStruct["wellPoissMouv"])):
    for j in range(0, len(superStruct["wellPoissMouv"][i])):
      curvatures = {}
      if hyperparameters["saveAllDataEvenIfNotInBouts"]:
        fname = os.path.join(hyperparameters["outputFolder"], videoNameWithTimestamp, f'allData_{hyperparameters["videoName"]}_wellNumber{i}_animal{j}.csv')
        startLines = []
        with open(fname) as f:
          line = f.readline()
          while ',' not in line:
            startLines.append(line)
            line = f.readline()
        df = pd.read_csv(fname, skiprows=len(startLines))
        df = df.loc[:, ~df.columns.str.contains('^Unnamed')]
      for k in range(0, len(superStruct["wellPoissMouv"][i][j])):
        
        # Creation of the curvature graph for bout k
        TailX_VideoReferential = superStruct["wellPoissMouv"][i][j][k]["TailX_VideoReferential"]
        TailY_VideoReferential = superStruct["wellPoissMouv"][i][j][k]["TailY_VideoReferential"]

        if hyperparameters["videoPixelSize"]:
          tailLenghtInPixels = np.sum([math.sqrt((TailX_VideoReferential[0][l] - TailX_VideoReferential[0][l+1])**2 + (TailY_VideoReferential[0][l] - TailY_VideoReferential[0][l+1])**2) for l in range(0, len(TailX_VideoReferential[0]) - 1)])

        curvature = calculateCurvature(TailX_VideoReferential, TailY_VideoReferential, hyperparameters)
        originalCurvature = np.flip(np.transpose(curvature), 0)

        if hyperparameters["saveAllDataEvenIfNotInBouts"]:
          bStart = superStruct["wellPoissMouv"][i][j][k]["BoutStart"]
          bEnd = superStruct["wellPoissMouv"][i][j][k]["BoutEnd"]
          curvatures[(bStart - firstFrame, bEnd + 1 - firstFrame)] = originalCurvature.tolist()

        rolling_window = hyperparameters["curvatureMedianFilterSmoothingWindow"]
        if rolling_window:
          curvature = ndimage.median_filter(curvature, size=rolling_window) # 2d median filter
          curvature = np.flip(np.transpose(curvature), 0)
        else:
          curvature = originalCurvature
        
        superStruct["wellPoissMouv"][i][j][k]["curvature"] = curvature.tolist()

        if hyperparameters["saveCurvaturePlots"]:
          fig = plt.figure(1)
          if hyperparameters["maxCurvatureValues"] == 0:
            maxx = max([max(abs(np.array(t))) for t in curvature])
          else:
            maxx = hyperparameters["maxCurvatureValues"]          
          if hyperparameters["curvatureXaxisNbFrames"] != 0:
            if len(curvature[0]) < hyperparameters["curvatureXaxisNbFrames"]:
              curvature2 = np.pad(curvature, ((0, 0), (0, hyperparameters["curvatureXaxisNbFrames"] - len(curvature[0]))), mode='constant')
            elif len(curvature[0]) > hyperparameters["curvatureXaxisNbFrames"]:
              curvature2 = curvature[:, 0:hyperparameters["curvatureXaxisNbFrames"]]
            else:
              curvature2 = curvature
          else:
            curvature2 = curvature
          plt.pcolor(curvature2, vmin=-maxx, vmax=maxx, cmap=hyperparameters["colorMapCurvature"])
          ax = fig.axes
          
          if hyperparameters["videoFPS"] and int(len(curvature2[0])/10):
            ax[0].set_xlabel('Second')
            plt.xticks([i for i in range(0, len(curvature2[0]), int(len(curvature2[0])/10))], [int(100*(i/hyperparameters["videoFPS"]))/100 for i in range(0, len(curvature2[0]), int(len(curvature2[0])/10))])
          else:
            ax[0].set_xlabel('Frame number')
            
          if hyperparameters["videoPixelSize"] and int(len(curvature2)/10):
            ax[0].set_ylabel('Rostral to Caudal (in mm)')
            plt.yticks([i for i in range(0, len(curvature2), int(len(curvature2)/10))], [int(100*( hyperparameters["videoPixelSize"] * tailLenghtInPixels * (i/len(curvature2)) ))/100 for i in range(0, len(curvature2), int(len(curvature2)/10))])
          else:
            ax[0].set_ylabel('Rostral to Caudal (arbitrary unit)')
          
          plt.colorbar()
          plt.savefig(os.path.join(outputPath, hyperparameters["videoName"] + "_curvature_bout" + str(i) + '_' + str(j) + '_' + str(k) + '.png'))
          plt.close(1)
        
        if hyperparameters["saveCurvatureData"]:
          curvatureMatrixFile = open(os.path.join(outputPath, hyperparameters["videoName"] + "_curvatureData" + str(i) + '_' + str(j) + '_' + str(k) + '.txt'), 'wb')
          pickle.dump(curvature, curvatureMatrixFile)
          curvatureMatrixFile.close()
        
        
        # Creation of tail angle graph for bout k
        if hyperparameters["saveTailAngleGraph"]:
          bStart = superStruct["wellPoissMouv"][i][j][k]["BoutStart"]
          bEnd   = superStruct["wellPoissMouv"][i][j][k]["BoutEnd"]
          tailAngle = superStruct["wellPoissMouv"][i][j][k]["TailAngle_smoothed"]
          if bEnd - bStart + 1 == len(tailAngle):
            plt.figure(1)
            plt.plot([i for i in range(bStart, bEnd + 1)], [t*(180/math.pi) for t in tailAngle])
            if hyperparameters["perBoutOutputYaxis"]:
              plt.axis([bStart, bEnd + 1, hyperparameters["perBoutOutputYaxis"][0], hyperparameters["perBoutOutputYaxis"][1]])
            plt.savefig(os.path.join(outputPath, hyperparameters["videoName"] + "_angle_bout" + str(i) + '_' + str(j) + '_' + str(k) + '.png'))
            plt.close(1)
        
        # Creation of sub-video for bout k
        if hyperparameters["saveSubVideo"]:
          x0 = superStruct["wellPoissMouv"][i][j][k]["HeadX"][0]
          y0 = superStruct["wellPoissMouv"][i][j][k]["HeadY"][0]
          TailX_VideoReferential = superStruct["wellPoissMouv"][i][j][k]["TailX_VideoReferential"]
          TailY_VideoReferential = superStruct["wellPoissMouv"][i][j][k]["TailY_VideoReferential"]
          tailExtX = TailX_VideoReferential[0][len(TailX_VideoReferential[0])-1]
          tailExtY = TailY_VideoReferential[0][len(TailY_VideoReferential[0])-1]
          dist = int((math.sqrt((tailExtX-x0)**2 + (tailExtY-y0)**2))*1.3)
          
          outputName = os.path.join(outputPath, hyperparameters["videoName"] + "_bout" + str(i) + '_' + str(j) + '_' + str(k) + '.avi')
          if dist < cap.get(3) and dist < cap.get(4):
            outputVideoX = int(dist)
            outputVideoY = int(dist)
          else:
            outputVideoX = min(cap.get(3), cap.get(4))
            outputVideoY = min(cap.get(3), cap.get(4))
          
          out = cv2.VideoWriter(outputName,cv2.VideoWriter_fourcc('M','J','P','G'), 10, (outputVideoX, outputVideoY))
          
          BoutStart = superStruct["wellPoissMouv"][i][j][k]["BoutStart"] - perBoutOutputVideoStartStopFrameMargin
          BoutEnd = superStruct["wellPoissMouv"][i][j][k]["BoutEnd"] + perBoutOutputVideoStartStopFrameMargin
          if BoutStart < 0:
            BoutStart = 0
          if BoutEnd >= lastFrame:
            BoutEnd = lastFrame - 1
          mouvLength = len(superStruct["wellPoissMouv"][i][j][k]["HeadX"])
          cap.set(1, BoutStart - firstFrame)
          
          for l in range(BoutStart, BoutEnd):
            ret, frame = cap.read()
            
            if l < superStruct["wellPoissMouv"][i][j][k]["BoutStart"]:
              l2 = 0
            elif l >= superStruct["wellPoissMouv"][i][j][k]["BoutEnd"]:
              l2 = mouvLength - 1
            else:
              l2 = l - superStruct["wellPoissMouv"][i][j][k]["BoutStart"]
            x = superStruct["wellPoissMouv"][i][j][k]["HeadX"][l2]
            y = superStruct["wellPoissMouv"][i][j][k]["HeadY"][l2]
            Heading = superStruct["wellPoissMouv"][i][j][k]["Heading"][l2]
            
            x = int(x + infoWells[i][0])
            y = int(y + infoWells[i][1])
            
            rows = len(frame)
            cols = len(frame[0])
            M = cv2.getRotationMatrix2D((x, y), -(((math.pi/2) - (Heading+math.pi))%(2*math.pi))*(180/math.pi), 1)
            frame = cv2.warpAffine(frame, M, (cols,rows))
            
            if int(x-dist/2)+outputVideoX <= 0:
              print("problem in perBoutOutput")
            else:
              if int(x-dist/2) > 0:
                frame = frame[int(y):int(y)+outputVideoY, int(x-dist/2):int(x-dist/2)+outputVideoX]
              else:
                frame = frame[int(y):int(y)+outputVideoY, 0:outputVideoX]
            
            frame2 = np.zeros((outputVideoX, outputVideoY, 3), np.uint8)
            frame2[0:len(frame), 0:len(frame[0]), :] = frame[0:len(frame), 0:len(frame[0]), :]
            frame = frame2
            
            cv2.putText(frame, str(l + hyperparameters["firstFrame"] - 1), (int(outputVideoX - 100), int(outputVideoY - 30)), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0))
            
            out.write(frame)
            
          out.release()

      if hyperparameters["saveAllDataEvenIfNotInBouts"]:
        curvatureCount = max(map(len, curvatures.values())) if curvatures else 0
        curvatureData = [[float('nan')] * nbFrames for _ in range(curvatureCount)]
        for (start, end), values in curvatures.items():
          for data, vals in zip(curvatureData, values):
            data[start:end] = vals
        for idx, data in enumerate(curvatureData):
          df[f'curvature{idx + 1}'] = data
        with open(fname, 'w+', newline='') as f:
          f.write(''.join(startLines))
          df.convert_dtypes().to_csv(f)
      if hyperparameters['storeH5'] and superStruct["wellPoissMouv"][i][j]:
        with h5py.File(hyperparameters['H5filename'], 'a') as results:
          dataGroup = results.require_group(f"dataForWell{i}/dataForAnimal{j}/dataPerFrame")
          curvature = np.empty((lastFrame - firstFrame + 1, hyperparameters['nbTailPoints'] - 2), dtype=float)
          TailX_VideoReferential = np.column_stack([dataGroup['HeadPos']['X']] + [dataGroup['TailPosX'][col] for col in dataGroup['TailPosX'].attrs['columns']])
          TailY_VideoReferential = np.column_stack([dataGroup['HeadPos']['Y']] + [dataGroup['TailPosY'][col] for col in dataGroup['TailPosY'].attrs['columns']])
          curvature = list(np.flip(np.transpose(calculateCurvature(TailX_VideoReferential, TailY_VideoReferential, hyperparameters)), 0))
          data = np.empty(len(curvature[0]), dtype=[(f'Pos{idx}', float) for idx in range(1, len(curvature) + 1)])
          for idx, curvatureData in enumerate(curvature):
            data[f'Pos{idx + 1}'] = curvatureData
          dataset = dataGroup.create_dataset('curvature', data=data)
          dataset.attrs['columns'] = data.dtype.names

  return superStruct



import numpy as np
import cv2
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
from zebrazoom.code.GUI.getCoordinates import findWellLeft, findWellRight, findHeadCenter, findBodyExtremity
from zebrazoom.code.GUI.adjustParameterInsideAlgo import adjustBoutDetectionOnlyPage
from zebrazoom.code.GUI.automaticallyFindOptimalParameters import automaticallyFindOptimalParameters

import math
from zebrazoom.code.findWells import findWells
from zebrazoom.code.getHyperparameters import getHyperparametersSimple
from zebrazoom.code.tracking import getBackground
import pickle
from zebrazoom.zebraZoomVideoAnalysis import ZebraZoomVideoAnalysis
from zebrazoom.code.tracking import get_default_tracking_method
import json
import os

from PyQt5.QtCore import Qt, QEventLoop, QTimer
from PyQt5.QtWidgets import QApplication, QFileDialog, QGridLayout, QLabel, QMessageBox, QVBoxLayout

import zebrazoom.code.paths as paths
import zebrazoom.code.util as util


_VIDEO_EXTENSIONS = {'.264', '.3g2', '.3gp', '.3gp2', '.3gpp', '.3gpp2', '.3mm', '.3p2', '.60d', '.787', '.89', '.aaf', '.aec', '.aep', '.aepx', '.aet', '.aetx', '.ajp',
                     '.ale', '.am', '.amc', '.amv', '.amx', '.anim', '.aqt', '.arcut', '.arf', '.asf', '.asx', '.avb', '.avc', '.avd', '.avi', '.avp', '.avs', '.avs',
                     '.avv', '.axm', '.bdm', '.bdmv', '.bdt2', '.bdt3', '.bik', '.bix', '.bmk', '.bnp', '.box', '.bs4', '.bsf', '.bvr', '.byu', '.camproj', '.camrec',
                     '.camv', '.ced', '.cel', '.cine', '.cip', '.clpi', '.cmmp', '.cmmtpl', '.cmproj', '.cmrec', '.cpi', '.cst', '.cvc', '.cx3', '.d2v', '.d3v', '.dat',
                     '.dav', '.dce', '.dck', '.dcr', '.dcr', '.ddat', '.dif', '.dir', '.divx', '.dlx', '.dmb', '.dmsd', '.dmsd3d', '.dmsm', '.dmsm3d', '.dmss', '.dmx', '.dnc',
                     '.dpa', '.dpg', '.dream', '.dsy', '.dv', '.dv-avi', '.dv4', '.dvdmedia', '.dvr', '.dvr-ms', '.dvx', '.dxr', '.dzm', '.dzp', '.dzt', '.edl', '.evo', '.eye',
                     '.ezt', '.f4p', '.f4v', '.fbr', '.fbr', '.fbz', '.fcp', '.fcproject', '.ffd', '.flc', '.flh', '.fli', '.flv', '.flx', '.gfp', '.gl', '.gom', '.grasp',
                     '.gts', '.gvi', '.gvp', '.h264', '.hdmov', '.hkm', '.ifo', '.imovieproj', '.imovieproject', '.ircp', '.irf', '.ism', '.ismc', '.ismv', '.iva', '.ivf',
                     '.ivr', '.ivs', '.izz', '.izzy', '.jss', '.jts', '.jtv', '.k3g', '.kmv', '.ktn', '.lrec', '.lsf', '.lsx', '.m15', '.m1pg', '.m1v', '.m21', '.m21', '.m2a',
                     '.m2p', '.m2t', '.m2ts', '.m2v', '.m4e', '.m4u', '.m4v', '.m75', '.mani', '.meta', '.mgv', '.mj2', '.mjp', '.mjpg', '.mk3d', '.mkv', '.mmv', '.mnv',
                     '.mob', '.mod', '.modd', '.moff', '.moi', '.moov', '.mov', '.movie', '.mp21', '.mp21', '.mp2v', '.mp4', '.mp4v', '.mpe', '.mpeg', '.mpeg1', '.mpeg4',
                     '.mpf', '.mpg', '.mpg2', '.mpgindex', '.mpl', '.mpl', '.mpls', '.mpsub', '.mpv', '.mpv2', '.mqv', '.msdvd', '.mse', '.msh', '.mswmm', '.mts', '.mtv',
                     '.mvb', '.mvc', '.mvd', '.mve', '.mvex', '.mvp', '.mvp', '.mvy', '.mxf', '.mxv', '.mys', '.ncor', '.nsv', '.nut', '.nuv', '.nvc', '.ogm', '.ogv', '.ogx',
                     '.osp', '.otrkey', '.pac', '.par', '.pds', '.pgi', '.photoshow', '.piv', '.pjs', '.playlist', '.plproj', '.pmf', '.pmv', '.pns', '.ppj', '.prel', '.pro',
                     '.prproj', '.prtl', '.psb', '.psh', '.pssd', '.pva', '.pvr', '.pxv', '.qt', '.qtch', '.qtindex', '.qtl', '.qtm', '.qtz', '.r3d', '.rcd', '.rcproject',
                     '.rdb', '.rec', '.rm', '.rmd', '.rmd', '.rmp', '.rms', '.rmv', '.rmvb', '.roq', '.rp', '.rsx', '.rts', '.rts', '.rum', '.rv', '.rvid', '.rvl', '.sbk',
                     '.sbt', '.scc', '.scm', '.scm', '.scn', '.screenflow', '.sec', '.sedprj', '.seq', '.sfd', '.sfvidcap', '.siv', '.smi', '.smi', '.smil', '.smk', '.sml',
                     '.smv', '.spl', '.sqz', '.srt', '.ssf', '.ssm', '.stl', '.str', '.stx', '.svi', '.swf', '.swi', '.swt', '.tda3mt', '.tdx', '.thp', '.tivo', '.tix',
                     '.tod', '.tp', '.tp0', '.tpd', '.tpr', '.trp', '.ts', '.tsp', '.ttxt', '.tvs', '.usf', '.usm', '.vc1', '.vcpf', '.vcr', '.vcv', '.vdo', '.vdr', '.vdx',
                     '.veg','.vem', '.vep', '.vf', '.vft', '.vfw', '.vfz', '.vgz', '.vid', '.video', '.viewlet', '.viv', '.vivo', '.vlab', '.vob', '.vp3', '.vp6', '.vp7',
                     '.vpj', '.vro', '.vs4', '.vse', '.vsp', '.w32', '.wcp', '.webm', '.wlmp', '.wm', '.wmd', '.wmmp', '.wmv', '.wmx', '.wot', '.wp3', '.wpl', '.wtv', '.wve',
                     '.wvx', '.xej', '.xel', '.xesc', '.xfl', '.xlmv', '.xmv', '.xvid', '.y4m', '.yog', '.yuv', '.zeg', '.zm1', '.zm2', '.zm3', '.zmv'}


def getMainArguments(self):
  s            = self.videoToCreateConfigFileFor
  arr          = s.split("/")
  nameWithExt  = arr.pop()
  pathToVideo  = '/'.join(arr) + '/'
  nameWithExtL = nameWithExt.split(".")
  videoExt     = nameWithExtL.pop()
  videoName    = '.'.join(nameWithExtL)
  configFile   = self.configFile
  argv         = ["outputFolder", self.ZZoutputLocation]
  return [pathToVideo, videoName, videoExt, configFile, argv]


def chooseVideoToCreateConfigFileFor(self, controller, reloadConfigFile):
  if int(reloadConfigFile):
    configFileName, _ =  QFileDialog.getOpenFileName(self.window, "Select configuration file", paths.getConfigurationFolder(), "JSON (*.json)")
    if not configFileName:
      return False
    try:
      with open(configFileName) as f:
        self.configFile = json.load(f)
    except (EnvironmentError, json.JSONDecodeError) as e:
      QMessageBox.critical(self.window, "Could not read config file", "Config file couldn't be read: %s\n" % str(e))
      return False
    if self.configFile.get('trackingImplementation') is not None:
      QMessageBox.information(self.window, "Optimization not supported", "Optimization is not yet supported for this type of configuration file.")
      self.configFile.clear()
      return False
    self.savedConfigFile = self.configFile.copy()

  self.videoToCreateConfigFileFor, _ = QFileDialog.getOpenFileName(self.window, "Select video to create config file for", os.path.expanduser("~"),
                                                                   filter=f'Videos ({" ".join("*%s" % ext for ext in _VIDEO_EXTENSIONS)});; All files (*.*)')
  if not self.videoToCreateConfigFileFor:
    self.configFile.clear()
    return False
  return True


@util.addToHistory
def chooseGeneralExperimentFirstStep(self, controller, freeZebra, headEmbZebra, drosophilia, rodent, other, fastScreen):
  if int(fastScreen):
    controller.show_frame("HomegeneousWellsLayout")
  else:
    self.configFile["extractAdvanceZebraParameters"] = 0
    if int(freeZebra):
      controller.show_frame("FreelySwimmingExperiment")
    else:
      chooseGeneralExperiment(self, controller, freeZebra, headEmbZebra, drosophilia, rodent, other, 0)


def chooseGeneralExperiment(self, controller, freeZebra, headEmbZebra, drosophilia, rodent, other, freeZebra2):
  self.configFile["extractAdvanceZebraParameters"] = 0
  if int(freeZebra):
    self.organism = 'zebrafish'
    self.configFile["headEmbeded"] = 0
    controller.show_frame("WellOrganisation")
  elif int(freeZebra2):
    self.organism = 'zebrafishNew'
    self.configFile["headEmbeded"] = 0
    controller.show_frame("WellOrganisation")
  elif int(headEmbZebra):
    self.organism = 'headembeddedzebrafish'
    self.configFile["headEmbeded"] = 1
    chooseBeginningAndEndOfVideo(self, controller)
  else:
    self.organism = 'drosoorrodent'
    self.configFile["headEmbeded"] = 0
    self.configFile["freeSwimmingTailTrackingMethod"] = "none"
    controller.show_frame("WellOrganisation")


def _getRegion(controller):
  cap = zzVideoReading.VideoCapture(controller.videoToCreateConfigFileFor)
  cap.set(1, 10)
  ret, frame = cap.read()
  back = False
  def backClicked():
    nonlocal back
    back = True
  rect = util.getRectangle(frame, "Click on the top left and bottom right of the region of interest", backBtnCb=backClicked)
  if back:
    controller.window.centralWidget().layout().setCurrentIndex(0)
    controller.configFileHistory[-2]()
    return None
  return rect


@util.addToHistory
def wellOrganisation(self, controller, circular, rectangular, roi, other, multipleROIs, groupSameSizeAndShapeEquallySpacedWells):
  if multipleROIs:
    self.shape = None
    controller.show_frame("NbRegionsOfInterest")
  else:
    if groupSameSizeAndShapeEquallySpacedWells:
      self.shape = 'groupSameSizeAndShapeEquallySpacedWells'
      self.configFile["groupOfMultipleSameSizeAndShapeEquallySpacedWells"] = 1
      controller.show_frame("CircularOrRectangularWells")
    else:
      if rectangular:
        self.shape = 'rectangular'
        self.configFile["wellsAreRectangles"] = 1
        controller.show_frame("CircularOrRectangularWells")
      else:
        if circular and self.organism != 'drosoorrodent': # should remove the self.organism != 'drosoorrodent' at some point
          self.shape = 'circular'
          controller.show_frame("CircularOrRectangularWells")
        else:
          self.shape = 'other'
          if roi:
            rect = _getRegion(controller)
            if rect is None:
              return
            self.configFile["oneWellManuallyChosenTopLeft"], self.configFile["oneWellManuallyChosenBottomRight"] = rect
            self.configFile["nbWells"] = 1
            util.addToHistory(chooseBeginningAndEndOfVideo)(self, controller)
          else:
            self.configFile["noWellDetection"] = 1
            self.configFile["nbWells"] = 1
            chooseBeginningAndEndOfVideo(self, controller)


@util.addToHistory
def regionsOfInterest(self, controller, nbwells):

  self.configFile["multipleROIsDefinedDuringExecution"] = 1
  self.configFile["nbWells"] = int(nbwells)

  chooseBeginningAndEndOfVideo(self, controller)


def rectangularWells(self, controller, nbwells, nbRowsOfWells, nbWellsPerRows):

  [pathToVideo, videoName, videoExt, configFile, argv] = getMainArguments(self)

  configFile["adjustRectangularWellsDetect"] = 1

  app = QApplication.instance()
  with app.busyCursor():
    try:
      storeH5 = configFile.get('storeH5')
      configFile['storeH5'] = 1
      ZebraZoomVideoAnalysis(pathToVideo, videoName, videoExt, configFile, argv).run()
    except ValueError:
      newhyperparameters = pickle.load(open(os.path.join(paths.getRootDataFolder(), 'newhyperparameters'), 'rb'))
      for index in newhyperparameters:
        configFile[index] = newhyperparameters[index]
    except NameError:
      print("Configuration file parameters changes discarded.")
    finally:
      if storeH5 is not None:
        configFile['storeH5'] = storeH5
      else:
        del configFile['storeH5']

  configFile["adjustRectangularWellsDetect"] = 0

  self.configFile = configFile

  chooseBeginningAndEndOfVideo(self, controller)


@util.addToHistory
def homegeneousWellsLayout(self, controller, nbRowsOfWells, nbWellsPerRows, detectBouts):

  self.configFile = {"trackingMethod": "fastCenterOfMassTracking_KNNbackgroundSubtraction", "nbAnimalsPerWell": 1, "nbWells": 8, "nbRowsOfWells": 2, "nbWellsPerRows": 4, "groupOfMultipleSameSizeAndShapeEquallySpacedWells": 1, "postProcessMultipleTrajectories": 0, "trackingPointSizeDisplay": 3, "extractAdvanceZebraParameters": 0,  "validationVideoPlotHeading": 0, "trackTail": 0, "freqAlgoPosFollow": 100, "fasterMultiprocessing": 1, "copyOriginalVideoToOutputFolderForValidation": 0, "backgroundSubtractorKNN": 1, "boutsMinNbFrames": 0, "addOneFrameAtTheEndForBoutDetection": 1, "fillGapFrameNb": 0} # "postProcessMultipleTrajectories": 1, "postProcessRemoveLowProbabilityDetection" : 1, "postProcessLowProbabilityDetectionPercentOfMaximum" : 0.2

  self.configFile["nbWells"] = int(nbRowsOfWells) * int(nbWellsPerRows)
  self.configFile["nbRowsOfWells"] = int(nbRowsOfWells)
  self.configFile["nbWellsPerRows"] = int(nbWellsPerRows)
  self.configFile["noBoutsDetection"] = int(not detectBouts)

  if detectBouts:
    controller.calculateBackgroundFreelySwim(controller, 0, boutDetectionsOnly=True)
  else:
    controller.show_frame("FinishConfig")


@util.addToHistory
def morePreciseFastScreen(self, controller, nbRowsOfWells, nbWellsPerRows, detectBouts):

  # The gaussian image filtering should be added here in the future
  self.configFile = {"trackingMethod": "fastCenterOfMassTracking_ClassicalBackgroundSubtraction", "minPixelDiffForBackExtract": 20, "backgroundPreProcessParameters": [[3]], "backgroundPreProcessMethod": ["erodeThenMin"], "trackingPointSizeDisplay": 1, "nbAnimalsPerWell": 1, "extractAdvanceZebraParameters": 0, "trackTail": 0, "nbWells": 1, "noWellDetection": 1, "backgroundExtractionForceUseAllVideoFrames": 1, "headSize": 2, "createValidationVideo": 0, "lastFrame": 1000}

  self.configFile["nbAnimalsPerWell"] = int(nbRowsOfWells) * int(nbWellsPerRows)

  # self.configFile["nbWells"]          = int(nbwells)

  self.configFile["nbRowsOfWells"] = int(nbRowsOfWells)
  self.configFile["nbWellsPerRows"] = int(nbWellsPerRows)
  self.configFile["noBoutsDetection"] = int(not detectBouts)

  self.calculateBackgroundFreelySwim(controller, 0, morePreciseFastScreen=True, boutDetectionsOnly=detectBouts)

  # controller.show_frame("FinishConfig")


@util.addToHistory
def circularOrRectangularWells(self, controller, nbRowsOfWells, nbWellsPerRows, nbanimals):
  self.configFile["nbWells"]        = int(nbRowsOfWells) * int(nbWellsPerRows)
  self.configFile["nbAnimalsPerWell"] = int(nbanimals)
  self.configFile["nbRowsOfWells"]  = int(nbRowsOfWells)
  self.configFile["nbWellsPerRows"]  = int(nbWellsPerRows)

  if self.shape == 'circular':
    controller.show_frame("ChooseCircularWellsLeft")
  else:
    if self.shape == 'groupSameSizeAndShapeEquallySpacedWells':
      chooseBeginningAndEndOfVideo(self, controller)
    else:
      rectangularWells(self, controller, self.configFile["nbWells"], nbRowsOfWells, nbWellsPerRows)


@util.addToHistory
def chooseCircularWellsLeft(self, controller):
  cap = zzVideoReading.VideoCapture(self.videoToCreateConfigFileFor)
  ret, frame = cap.read()
  wellLeft = findWellLeft(frame)
  if wellLeft is None:
    return
  [x, y] = wellLeft
  self.wellLeftBorderX = x
  self.wellLeftBorderY = y
  util.addToHistory(controller.show_frame)("ChooseCircularWellsRight")

@util.addToHistory
def chooseCircularWellsRight(self, controller):
  cap = zzVideoReading.VideoCapture(self.videoToCreateConfigFileFor)
  ret, frame = cap.read()
  wellRight = findWellRight(frame)
  if wellRight is None:
    return
  [xRight, yRight] = wellRight
  xLeft = self.wellLeftBorderX
  yLeft = self.wellLeftBorderY
  dist = math.sqrt((xLeft - xRight)**2 + (yLeft - yRight)**2)
  self.configFile["minWellDistanceForWellDetection"] = int(dist)
  self.configFile["wellOutputVideoDiameter"]         = int(dist + dist * 0.2)
  util.addToHistory(chooseBeginningAndEndOfVideo)(self, controller)


@util.addToHistory
def _beginningAndEndChosen(self, controller):
  controller.window.centralWidget().layout().setCurrentIndex(0)
  if int(controller.configFile["headEmbeded"]) == 1:
    controller.show_frame("HeadEmbeded")
  else:
    if controller.organism == 'zebrafishNew':
      controller.show_frame("NumberOfAnimals2")
    elif controller.organism == 'zebrafish':
      controller.show_frame("NumberOfAnimals")
    else:
      self.configFile["trackingMethod"] = "classicCenterOfMassTracking"
      controller.show_frame("NumberOfAnimalsCenterOfMass")


@util.addToHistory
def _chooseEnd(self, controller):
  util.chooseEndPage(controller, controller.videoToCreateConfigFileFor, "Choose where the analysis of your video should end.", "Ok, I want the tracking to end at this frame!", lambda: _beginningAndEndChosen(self, controller))


def chooseBeginningAndEndOfVideo(self, controller):
  util.chooseBeginningPage(controller, controller.videoToCreateConfigFileFor, "Choose where the analysis of your video should start.", "Ok, I want the tracking to start at this frame!", lambda: _chooseEnd(self, controller),
                           extraButtonInfo=("I want the tracking to run on the entire video!", lambda: _beginningAndEndChosen(self, controller), {"background_color": util.DEFAULT_BUTTON_COLOR}))


def getImageForMultipleAnimalGUI(l, nx, ny, max_l, videoToCreateConfigFileFor, background, wellPositions, hyperparameters):

  [frame, a1, a2] = get_default_tracking_method()(videoToCreateConfigFileFor, [], hyperparameters).getForegroundImage(background, l, 0)

  lengthX = nx * 2
  lengthY = ny

  frame2 = frame
  ret,thresh2 = cv2.threshold(frame2,hyperparameters["thresholdForBlobImg"],255,cv2.THRESH_BINARY)
  kernel  = np.ones((hyperparameters["erodeSize"],hyperparameters["erodeSize"]), np.uint8)
  thresh2 = cv2.dilate(thresh2, kernel, iterations=hyperparameters["dilateIter"])

  thresh = thresh2
  thresh2 = cv2.cvtColor(thresh2, cv2.COLOR_GRAY2RGB)
  frame   = cv2.cvtColor(frame,   cv2.COLOR_GRAY2RGB)
  thresh[0, :]                = 255
  thresh[:, 0]                = 255
  thresh[len(thresh)-1, :]    = 255
  thresh[:, len(thresh[0])-1] = 255
  areaList = []
  contours, hierarchy = cv2.findContours(thresh,cv2.RETR_TREE,cv2.CHAIN_APPROX_SIMPLE)
  for contour in contours:
    area = cv2.contourArea(contour)
    if area < (len(thresh) * len(thresh[0]))/2:
      areaList.append(area)
    if (area > hyperparameters["minArea"]) and (area < hyperparameters["maxArea"]):
      M = cv2.moments(contour)
      if M['m00']:
        x = int(M['m10']/M['m00'])
        y = int(M['m01']/M['m00'])
        cv2.circle(thresh2, (x,y), 3, (0,0,255), -1)
      else:
        x = 0
        y = 0

  frame = cv2.line(frame, (len(frame[0])-5, 0), (len(frame[0])-5, len(frame)), (255, 0, 0), 5)

  frame = np.concatenate((frame, thresh2), axis=1)

  if len(areaList):
    maxToReturn = int((max(areaList)+2)*2)
  else:
    maxToReturn = (len(thresh) * len(thresh[0]))/5

  return [frame, maxToReturn]


def _createWidget(layout, values, key, minn, maxx, name, updateFrame):
  slider = util.SliderWithSpinbox(values[key], minn, maxx, name=name)

  def valueChanged():
    values[key] = slider.value()
    updateFrame()
  slider.valueChanged.connect(valueChanged)

  if name != "Frame number":
    elements = layout.count() - 2  # frame, frameSlider
    row = elements // 2 + 2
    col = elements % 2
    layout.addWidget(slider, row, col, Qt.AlignmentFlag.AlignLeft if col else Qt.AlignmentFlag.AlignRight)
  else:
    layout.addWidget(slider, 1, 0, 1, 2, Qt.AlignmentFlag.AlignCenter)

  return slider


def identifyMultipleHead(self, controller, nbanimals):
  self.configFile["videoName"] = "configFilePrep"

  tempConfig = self.configFile

  app = QApplication.instance()
  with app.busyCursor():
    # Getting hyperparameters, wellPositions, and background
    hyperparameters = getHyperparametersSimple(tempConfig)
    wellPositions = util.addToHistory(findWells)(self.videoToCreateConfigFileFor, hyperparameters)
    background    = getBackground(self.videoToCreateConfigFileFor, hyperparameters)

  cur_dir_path = os.path.dirname(os.path.realpath(__file__))

  def imagesGenerator():
    images = (cv2.imread(os.path.join(cur_dir_path, 'no1.png')),
              cv2.imread(os.path.join(cur_dir_path, 'no2.png')),
              cv2.imread(os.path.join(cur_dir_path, 'ok1.png')))
    while True:
      yield from images
  images = imagesGenerator()

  label = QLabel()
  label.setMinimumSize(1, 1)
  layout = QVBoxLayout()
  layout.addWidget(label, alignment=Qt.AlignmentFlag.AlignCenter)
  timer = QTimer()
  timer.setInterval(2000)
  timer.timeout.connect(lambda: util.setPixmapFromCv(next(images), label))
  timer.start()
  util.showBlockingPage(layout, title="", buttons=(("Ok, I understand!", lambda: timer.stop()),), labelInfo=(next(images), label))

  # Manual parameters adjustements
  cap        = zzVideoReading.VideoCapture(self.videoToCreateConfigFileFor)
  nx         = int(cap.get(3))
  ny         = int(cap.get(4))
  max_l      = int(cap.get(7))

  hyperparameters["minArea"] = 5
  hyperparameters["maxArea"] = 800

  frameNum = {"frameNum": hyperparameters.get("firstFrame", 1)}
  minPixelDiffForBackExtract = [hyperparameters["minPixelDiffForBackExtract"]]
  thresholdForBlobImg        = [hyperparameters["thresholdForBlobImg"]]
  dilateIter                 = [hyperparameters["dilateIter"]]
  minArea                    = [hyperparameters["minArea"]]
  maxArea                    = [hyperparameters["maxArea"]]
  firstFrame = hyperparameters["firstFrame"] if "firstFrame" in hyperparameters else 1
  lastFrame  = hyperparameters["lastFrame"]-1 if "lastFrame" in hyperparameters else max_l - 10

  frame, maxAreaBlobs = getImageForMultipleAnimalGUI(frameNum["frameNum"], nx, ny, max_l, self.videoToCreateConfigFileFor, background, wellPositions, hyperparameters)

  label = QLabel()
  label.setMinimumSize(1, 1)
  layout = QGridLayout()
  layout.addWidget(label, 0, 0, 1, 2, Qt.AlignmentFlag.AlignCenter)
  layout.setRowStretch(0, 1)

  def updateFrame():
    frame, maxAreaBlobs = getImageForMultipleAnimalGUI(frameNum["frameNum"], nx, ny, max_l, self.videoToCreateConfigFileFor, background, wellPositions, hyperparameters)
    util.setPixmapFromCv(frame, label)
    minAreaWidget.setMaximum(maxAreaBlobs)
    maxAreaWidget.setMaximum(maxAreaBlobs)

  _createWidget(layout, frameNum, "frameNum", firstFrame, lastFrame, "Frame number", updateFrame)
  _createWidget(layout, hyperparameters, "minPixelDiffForBackExtract", 0, 255, "Threshold left image", updateFrame)
  _createWidget(layout, hyperparameters, "thresholdForBlobImg", 0, 255, "Threshold right image", updateFrame)
  _createWidget(layout, hyperparameters, "dilateIter", 0, 15, "Area dilatation", updateFrame)
  minAreaWidget = _createWidget(layout, hyperparameters, "minArea", 0, maxAreaBlobs, "Minimum area", updateFrame)
  maxAreaWidget = _createWidget(layout, hyperparameters, "maxArea", 0, maxAreaBlobs, "Maximum area", updateFrame)

  util.showBlockingPage(layout, title="Adjust Parameters: As much as possible, you must see red points on and only on animals on the right image.", buttons=(("Ok, done!", None),), labelInfo=(frame, label))

  del self.configFile["videoName"]

  self.configFile["minPixelDiffForBackExtract"] = int(hyperparameters["minPixelDiffForBackExtract"])
  self.configFile["thresholdForBlobImg"]        = int(hyperparameters["thresholdForBlobImg"])
  self.configFile["dilateIter"]                 = int(hyperparameters["dilateIter"])
  self.configFile["minArea"]                    = int(hyperparameters["minArea"])
  self.configFile["maxArea"]                    = int(hyperparameters["maxArea"])
  self.configFile["headSize"]        = math.sqrt((int(hyperparameters["minArea"]) + int(hyperparameters["maxArea"])) / 2)


def _calculateMaxDepthParams(app, frame):
  back = False

  def backClicked():
    nonlocal back
    back = True

  while 1:
    headCoordinates = util.getPoint(frame, "Click on center of the head (defined here by middle point between eyes and swim bladder)", zoomable=True, backBtnCb=backClicked)
    if back:
      app.window.centralWidget().layout().setCurrentIndex(0)
      app.configFileHistory[-2]()
      return None

    tailCoordinates = util.getPoint(cv2.circle(frame.copy(), tuple(headCoordinates), 2, (0, 0, 255), -1), "Click on the tip of the tail of the same fish", zoomable=True, backBtnCb=backClicked)
    if not back:
      return math.dist(headCoordinates, tailCoordinates)
    back = False


def _setFastTrackingParams(app):
  app.configFile["trackingImplementation"] = "fastFishTracking.tracking"
  app.configFile["detectMovementWithRawVideoInsideTracking"] = 0
  app.configFile["debugHeadEmbededFindNextPoints"] = 0
  app.configFile["backgroundExtractionWithOnlyTwoFrames"] = 1
  app.configFile["authorizedRelativeLengthTailEnd"] = 0.6
  app.configFile["thetaDiffAcceptAfterAuthorizedRelativeLengthTailEnd"] = 1.6
  app.configFile["nbList"] = 20
  app.configFile["nbListAfterAuthorizedRelativeLengthTailEnd"] = 25
  app.configFile["thetaDiffAccept"] = 1.6
  app.configFile["authorizedRelativeLengthTailEnd2"] = 0.8
  app.configFile["thetaDiffAcceptAfterAuthorizedRelativeLengthTailEnd2"] = 0.6
  app.configFile["nbListAfterAuthorizedRelativeLengthTailEnd2"] = 25
  app.configFile["maximumMedianValueOfAllPointsAlongTheTail"] = 250
  app.configFile["headEmbededParamTailDescentPixThreshStop"] = 250
  app.configFile["minimumHeadPixelValue"] = 240
  app.configFile["nbTailPoints"] = 6
  app.configFile["backgroundSubtractionOnWholeImage"] = 0
  app.configFile["backgroundSubtractionOnROIhalfDiameter"] = 90
  app.configFile["chooseWellsToRunTrackingOnWithFirstAndLastFrame"] = 0


def _setMaxDepthParams(app):
  frame = util.chooseFrame(app, app.videoToCreateConfigFileFor, "Choose a frame where the fish tail is straight (not swimming)", "Next")
  if frame is None:
    app.window.centralWidget().layout().setCurrentIndex(0)
    app.configFileHistory[-2]()
    return False
  maxDepth = _calculateMaxDepthParams(app, frame)
  if maxDepth is None:
    return False
  app.configFile["maxDepth"] = maxDepth
  app.configFile["steps"] = [maxDepth / 3.6, maxDepth / 2.4, maxDepth / 1.8]
  app.configFile["paramGaussianBlur"] = int(maxDepth / 2.4)
  if app.configFile["paramGaussianBlur"] % 2 == 0:
    app.configFile["paramGaussianBlur"] += 1
  return True


@util.addToHistory
def _fastTrackingBoutDetection(app, detectBoutsMethod):
  _setFastTrackingParams(app)
  if not detectBoutsMethod:
    app.configFile["detectBouts"] = 0
    app.configFile["thresForDetectMovementWithRawVideo"] = 0
    app.show_frame("FinishConfig")
    return
  videoName, videoExt = os.path.splitext(os.path.basename(app.videoToCreateConfigFileFor))
  videoExt = videoExt.lstrip('.')
  pathToVideo = os.path.dirname(app.videoToCreateConfigFileFor)
  app.configFile["exitAfterWellsDetection"] = 1
  app.wellPositions = []
  try:
    storeH5 = app.configFile.get('storeH5')
    app.configFile['storeH5'] = 1
    with app.busyCursor():
      ZebraZoomVideoAnalysis(pathToVideo, videoName, videoExt, app.configFile, ['outputFolder', app.ZZoutputLocation]).run()
  except ValueError:
      pass
  finally:
    del app.configFile["exitAfterWellsDetection"]
    if storeH5 is not None:
      app.configFile['storeH5'] = storeH5
    else:
      del app.configFile['storeH5']
  app.configFile["noBoutsDetection"] = 0
  if detectBoutsMethod == 1:
    app.configFile["detectMovementWithRawVideoInsideTracking"] = 1
  if detectBoutsMethod == 2:
    app.configFile["coordinatesOnlyBoutDetection"] = 1
  adjustBoutDetectionOnlyPage(useBackground=False)


@util.addToHistory
def numberOfAnimals(nbanimals, animalsAlwaysVisible, forceBlobMethodForHeadTracking, detectBoutsMethod, algorithm, calculateBends, adjustBackgroundExtractionBasedOnNumberOfBlackPixels):
  app = QApplication.instance()

  app.configFile["noBoutsDetection"] = 1
  app.configFile["noChecksForBoutSelectionInExtractParams"] = 1
  app.configFile["trackingPointSizeDisplay"] = 4
  app.configFile["validationVideoPlotHeading"] = 0

  if calculateBends:
    app.configFile["extractAdvanceZebraParameters"] = 1

  nbanimals = int(nbanimals) if nbanimals is not None else app.configFile["nbWells"] * app.configFile["nbAnimalsPerWell"]
  if nbanimals == app.configFile["nbWells"]:
    app.configFile["nbAnimalsPerWell"] = 1
  else:
    app.configFile["nbAnimalsPerWell"] = int(nbanimals / app.configFile["nbWells"])
    app.configFile["multipleHeadTrackingIterativelyRelaxAreaCriteria"] = int(animalsAlwaysVisible)

  app.forceBlobMethodForHeadTracking = int(forceBlobMethodForHeadTracking)
  if app.forceBlobMethodForHeadTracking:
    app.configFile["forceBlobMethodForHeadTracking"] = app.forceBlobMethodForHeadTracking

  if algorithm == 2:
    if not _setMaxDepthParams(app):
      return
    _fastTrackingBoutDetection(app, detectBoutsMethod)
    if app.configFile["nbWells"] > 1:
      del app.configFile["backgroundSubtractionOnROIhalfDiameter"]
      app.configFile["backgroundSubtractionOnWholeImage"] = 1
  elif app.organism == 'zebrafish':
    app.show_frame("IdentifyHeadCenter")
  elif app.organism == 'zebrafishNew':
    automaticallyFindOptimalParameters(app, app, True, detectBoutsMethod, algorithm == 1, not animalsAlwaysVisible, adjustBackgroundExtractionBasedOnNumberOfBlackPixels)
    app.window.centralWidget().layout().currentWidget().refreshPage(showFasterTracking=algorithm == 1)
  elif app.organism == 'drosoorrodent' and not algorithm:
    automaticallyFindOptimalParameters(app, app, True, 0, False, not animalsAlwaysVisible, adjustBackgroundExtractionBasedOnNumberOfBlackPixels)
    app.window.centralWidget().layout().currentWidget().refreshPage(showFasterTracking=algorithm == 1)
  else:
    identifyMultipleHead(app, app, nbanimals)
    util.addToHistory(app.show_frame)("FinishConfig")
    app.window.centralWidget().layout().currentWidget().refreshPage(showFasterTracking=algorithm == 1)


@util.addToHistory
def chooseHeadCenter(self, controller):
  cap = zzVideoReading.VideoCapture(self.videoToCreateConfigFileFor)
  ret, frame = cap.read()
  headCenter = findHeadCenter(frame)
  if headCenter is None:
    return
  [x, y] = headCenter
  self.headCenterX = x
  self.headCenterY = y
  util.addToHistory(controller.show_frame)("IdentifyBodyExtremity")


@util.addToHistory
def chooseBodyExtremity(self, controller):
  cap = zzVideoReading.VideoCapture(self.videoToCreateConfigFileFor)
  ret, frame = cap.read()
  bodyExtremity = findBodyExtremity(frame)
  if bodyExtremity is None:
    return
  [extX, extY] = bodyExtremity
  headCenterX = self.headCenterX
  headCenterY = self.headCenterY
  dist = math.sqrt((extX - headCenterX)**2 + (extY - headCenterY)**2)

  if self.organism == 'zebrafish':
    minArea = int(dist * (dist * 0.1))
    maxArea = int(dist * (dist * 0.4))
    self.configFile["minArea"]     = minArea
    self.configFile["maxArea"]     = maxArea
    self.configFile["minAreaBody"] = minArea
    self.configFile["maxAreaBody"] = maxArea
    self.configFile["headSize"]    = int(dist * 0.3)
    self.configFile["minTailSize"] = int(dist * 0.05)
    self.configFile["maxTailSize"] = int(dist * 2)
    self.configFile["paramGaussianBlur"] = int(int(dist*(31/52))/2)*2 + 1
  else:
    minArea = int(((2 * dist) * (2 * dist)) * 0.2)
    maxArea = int(((2 * dist) * (2 * dist)) * 1.5)
    self.configFile["minArea"]     = minArea
    self.configFile["maxArea"]     = maxArea
    self.configFile["minAreaBody"] = minArea
    self.configFile["maxAreaBody"] = maxArea
    self.configFile["headSize"]    = int(dist * 2)

  self.configFile["extractBackWhiteBackground"] = 1

  self.configFile["noBoutsDetection"] = 1

  if int(self.configFile["nbAnimalsPerWell"]) > 1 or self.forceBlobMethodForHeadTracking:
    identifyMultipleHead(self, controller, int(self.configFile["nbAnimalsPerWell"]))

  util.addToHistory(controller.show_frame)("GoToAdvanceSettings")


@util.addToHistory
def goToAdvanceSettings(self, controller, yes, no):

  if int(no):
    controller.show_frame("FinishConfig")
  else:
    self.configFile["noBoutsDetection"] = 0
    self.calculateBackgroundFreelySwim(controller, 0)



import zebrazoom.code.tracking
import math
import cv2

class ExampleDummyHeadAndTailTracking(zebrazoom.code.tracking.BaseTrackingMethod):
  
  def __init__(self, videoPath, wellPositions, hyperparameters):
    self._videoPath = videoPath
    self._wellPositions = wellPositions
    self._hyperparameters = hyperparameters

  def run(self):
    
    ### Step 1 (out of 2): Tracking:
    
    # Getting video reader
    cap = cv2.VideoCapture(self._videoPath)
    if (cap.isOpened()== False):
      print("Error opening video stream or file")
    
    # Simple background extraction with first and last frame of the video
    ret, self._background = cap.read()
    cap.set(cv2.CAP_PROP_POS_FRAMES, int(cap.get(cv2.CAP_PROP_FRAME_COUNT)) - 1)
    ret, frame = cap.read()
    self._background = cv2.max(frame, self._background)
    self._background = cv2.cvtColor(self._background, cv2.COLOR_BGR2GRAY)
    cap.set(cv2.CAP_PROP_POS_FRAMES, 0)
    
    # Initializing variables
    ret = True
    trackingDataPerWell = {}
    for wellNumber in range(0, len(self._wellPositions)):
      trackingDataPerWell[wellNumber] = []
    
    # Going through each frame of the video
    while (ret):
      ret, frame = cap.read()
      if ret:
      
        # Subtracting background of image
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        putToWhite = ( frame.astype('int32') >= (self._background.astype('int32') - self._hyperparameters["minPixelDiffForBackExtract"]) )
        frame[putToWhite] = 255
      
        # Going through each well/arena/tank and applying tracking method on it
        for wellNumber in range(0, len(self._wellPositions)):
          
          # Retrieving well/tank/arena coordinates and selecting ROI
          wellXtop = self._wellPositions[wellNumber]['topLeftX']
          wellYtop = self._wellPositions[wellNumber]['topLeftY']
          lenghtWell_X = self._wellPositions[wellNumber]['lengthX']
          lenghtWell_Y = self._wellPositions[wellNumber]['lengthY']
          frameROI = frame[wellYtop:wellYtop+lenghtWell_Y, wellXtop:wellXtop+lenghtWell_X].copy()
          
          # Applying gaussian filter to find the location of the animal
          paramGaussianBlur = 35
          frameROI = cv2.GaussianBlur(frameROI, (paramGaussianBlur, paramGaussianBlur), 0)
          (minVal, maxVal, headPosition, maxLoc) = cv2.minMaxLoc(frameROI)
          
          # Storing track data for current well and current frame
          trackingDataPerWell[wellNumber].append(headPosition)
    
    
    ### Step 2 (out of 2): Extracting bout of movements:
    
    outputData = {} # Each element of this object will correspond to the tracking data of a particular well/tank/arena
    
    for wellNumber in range(0, len(self._wellPositions)):
      
      outputDataForWell = []
        
      if self._hyperparameters["detectBouts"]: # See below ("else") for no bout detection scenario
        
        nbFramesStepToAvoidNoise = 10
        minNbFramesForBoutDetect = 10
        
        # Finding frames with an instantaneous distance over a predifined threshold
        boutOccuring = [math.sqrt((trackingDataPerWell[wellNumber][i+nbFramesStepToAvoidNoise][0] - trackingDataPerWell[wellNumber][i][0])**2 + (trackingDataPerWell[wellNumber][i+nbFramesStepToAvoidNoise][1] - trackingDataPerWell[wellNumber][i][1])**2) > self._hyperparameters["minimumInstantaneousDistanceForBoutDetect"] for i in range(0, len(trackingDataPerWell[wellNumber])-nbFramesStepToAvoidNoise)]
        
        # Detecting bouts by finding long enough sequence of frames with high enough instantaneous distance
        boutCurrentlyOccuring = False
        boutFrameNumberStart  = -1
        for frameNumber, boutIsOccuring in enumerate(boutOccuring):
          if boutIsOccuring:
            if not(boutCurrentlyOccuring):
              boutFrameNumberStart = frameNumber
            boutCurrentlyOccuring = True
          else:
            boutCurrentlyOccuring = False
            if frameNumber - boutFrameNumberStart > minNbFramesForBoutDetect:
              # Saving information for each bout of movement detected
              boutOfMovement = {}
              boutOfMovement["AnimalNumber"]  = 0
              boutOfMovement["BoutStart"]     = boutFrameNumberStart
              boutOfMovement["BoutEnd"]       = frameNumber
              boutOfMovement["HeadX"]         = [trackingDataPerWell[wellNumber][i][0] for i in range(boutFrameNumberStart, frameNumber+1)]
              boutOfMovement["HeadY"]         = [trackingDataPerWell[wellNumber][i][1] for i in range(boutFrameNumberStart, frameNumber+1)]
              
              boutOfMovement["Heading"]                = [0.4 for i in range(boutFrameNumberStart, frameNumber+1)]
              boutOfMovement["TailAngle_Raw"]          = [math.cos(i/2) for i in range(boutFrameNumberStart, frameNumber+1)]
              boutOfMovement["TailX_VideoReferential"] = [[trackingDataPerWell[wellNumber][i][0] + j * 3 for j in range(0, 10)] for i in range(boutFrameNumberStart, frameNumber+1)]
              boutOfMovement["TailY_VideoReferential"] = [[trackingDataPerWell[wellNumber][i][1] + j * 3 for j in range(0, 10)] for i in range(boutFrameNumberStart, frameNumber+1)]
              
              boutOfMovement["leftEyeX"]      = [trackingDataPerWell[wellNumber][i][0] - 20 for i in range(boutFrameNumberStart, frameNumber+1)]
              boutOfMovement["leftEyeY"]      = [trackingDataPerWell[wellNumber][i][1] - 20 for i in range(boutFrameNumberStart, frameNumber+1)]
              boutOfMovement["leftEyeAngle"]  = [trackingDataPerWell[wellNumber][i][0] - 10 for i in range(boutFrameNumberStart, frameNumber+1)]
              boutOfMovement["leftEyeArea"]   = [trackingDataPerWell[wellNumber][i][0] - 10 for i in range(boutFrameNumberStart, frameNumber+1)]
              boutOfMovement["rightEyeX"]     = [trackingDataPerWell[wellNumber][i][0] - 20 for i in range(boutFrameNumberStart, frameNumber+1)]
              boutOfMovement["rightEyeY"]     = [trackingDataPerWell[wellNumber][i][1] + 20 for i in range(boutFrameNumberStart, frameNumber+1)]
              boutOfMovement["rightEyeAngle"] = [trackingDataPerWell[wellNumber][i][0] - 10 for i in range(boutFrameNumberStart, frameNumber+1)]
              boutOfMovement["rightEyeArea"]  = [trackingDataPerWell[wellNumber][i][0] - 10 for i in range(boutFrameNumberStart, frameNumber+1)]
              
              # Other keys can be added to "boutOfMovement", such as: Heading, TailAngle_Raw, TailX_VideoReferential, TailY_VideoReferential (see documentation)
              # You may also add: leftEyeX, leftEyeY, leftEyeAngle, leftEyeArea, rightEyeX, rightEyeY, rightEyeAngle, rightEyeArea: please contact us if you would like to track the eyes
              outputDataForWell.append(boutOfMovement)
      
      # Saving all tracked bouts of movements for current frame
      outputData[wellNumber] = outputDataForWell
    
    return outputData


zebrazoom.code.tracking.register_tracking_method('examples.exampleDummyHeadAndTailTracking', ExampleDummyHeadAndTailTracking)



from .createDistanceBetweenFramesExcelFile import *
from .getCurvaturePerBout import *
from .getCurvaturePerTimeInterval import *
from .getDataPerBout import *
from .getDataPerTimeInterval import *
from .getFirstAndLastFrame import *
from .getKinematicParametersPerBout import *
from .getTailAngleHeatmapPerBout import *
from .getTailAngleHeatmapPerTimeInterval import *
from .listAllBouts import *
from .plotCurvatureYaxisExact import *
from .plotCurvatureYaxisApproximate import *
from .plotTailAngleHeatmap import *
from .setFPSandPixelSize import *
from .getFPSandPixelSize import *
from .applyMedianFilterOnCurvature import *



import numpy as np

from zebrazoom.code.dataPostProcessing.perBoutOutput import calculateCurvature
from zebrazoom.code.getHyperparameters import getHyperparametersSimple


def calculateAndStoreCurvature(results, dataGroup):
  hyperparameters = getHyperparametersSimple(dict(results['configurationFileUsed'].attrs))
  firstFrame = results.attrs['firstFrame']
  lastFrame = results.attrs['lastFrame']
  curvature = np.empty((lastFrame - firstFrame + 1, hyperparameters['nbTailPoints'] - 2), dtype=float)
  TailX_VideoReferential = np.column_stack([dataGroup['HeadPos']['X']] + [dataGroup['TailPosX'][col] for col in dataGroup['TailPosX'].attrs['columns']])
  TailY_VideoReferential = np.column_stack([dataGroup['HeadPos']['Y']] + [dataGroup['TailPosY'][col] for col in dataGroup['TailPosY'].attrs['columns']])
  curvature = list(np.flip(np.transpose(calculateCurvature(TailX_VideoReferential, TailY_VideoReferential, hyperparameters)), 0))
  data = np.empty(len(curvature[0]), dtype=[(f'Pos{idx}', float) for idx in range(1, len(curvature) + 1)])
  for idx, curvatureData in enumerate(curvature):
    data[f'Pos{idx + 1}'] = curvatureData
  dataset = dataGroup.create_dataset('curvature', data=data)
  dataset.attrs['columns'] = data.dtype.names
  return curvature



import h5py
import numpy as np
import math
import os
import sys
import pandas as pd
from scipy.interpolate import UnivariateSpline

import numpy as np
# from filterpy.kalman import KalmanFilter
# from filterpy.common import Q_discrete_white_noise

from zebrazoom.code.detectMovementWithRawVideo import detectMovementWithRawVideo


def distBetweenThetas(theta1, theta2):
  diff = 0
  if theta1 > theta2:
    diff = theta1 - theta2
  else:
    diff = theta2 - theta1
  if diff > math.pi:
    diff = (2 * math.pi) - diff
  return diff

def calculateAngle(vectStart, vectEnd):
  x = vectEnd[0] - vectStart[0]
  y = vectEnd[1] - vectStart[1]
  if x == 0:
    if y > 0:
      heading = math.pi / 2
    else:
      heading = (3 * math.pi) / 2
  else:
    heading = np.arctan(abs(y/x))
    if (x < 0) and (y > 0):
      heading = math.pi - heading
    elif (x < 0) and (y < 0):
      heading = heading + math.pi
    elif (x < 0) and (y == 0):
      heading = math.pi
    elif (x > 0) and (y < 0):
      heading = 2*math.pi - heading
  return heading
  
def calculateTailAngle(angle1, angle2):
  output = angle1 - angle2
  output = (output + 2 * 3.14159265) % (2 * 3.14159265)
  if output > 3.14159265:
    output = output - 2*3.14159265;
  return output
  
def smoothAllTailAngles(allAngles, hyperparameters, start, end):
  # The first angle is removed here because it corresponds to the angle between two same point (the center of the head)
  tailangles_arr = np.transpose(allAngles[start:end+1, 1:len(allAngles)])
  tailangles_arr_smoothed = np.zeros((0, len(tailangles_arr[0])))
  for angle_raw in tailangles_arr:
    rolling_window = hyperparameters["tailAngleMedianFilter"]
    if rolling_window > 0:
      shift = int(-rolling_window / 2)
      angle_median = np.array(pd.Series(angle_raw).rolling(rolling_window).median())
      angle_median = np.roll(angle_median, shift)
      for ii in range(0, rolling_window):
        if ii >= 0 and ii < len(angle_raw):
          angle_median[ii] = angle_raw[ii]
      for ii in range(len(angle_median)-rolling_window,len(angle_median)):
        if ii >= 0 and ii < len(angle_raw):
          angle_median[ii] = angle_raw[ii]
    else:
      angle_median = angle_raw
    tailToSmooth = angle_median
    if len(tailToSmooth) >= 5:
      x = np.linspace(0, 1, len(tailToSmooth))
      s = UnivariateSpline(x, tailToSmooth, s=hyperparameters["tailAngleSmoothingFactor"])
      tailSmoothed     = s(x)
    else:
      tailSmoothed = tailToSmooth
      
    tailSmoothed2    = np.zeros((1, len(tailSmoothed)))
    tailSmoothed2[0] = tailSmoothed
      
    tailangles_arr_smoothed = np.append(tailangles_arr_smoothed, tailSmoothed2, axis=0)
  return [tailangles_arr, tailangles_arr_smoothed]

def _calculateTailLength(tailX, tailY):
  return np.sum([math.sqrt((tailX[x] - tailX[x+1]) ** 2 + (tailY[x] - tailY[x+1]) ** 2) for x in range(len(tailX) - 1)])

def extractParameters(trackingData, wellNumber, hyperparameters, videoPath, wellPositions, background, tailAngle = 0):

  firstFrame = hyperparameters["firstFrame"]
  thresAngleBoutDetect = hyperparameters["thresAngleBoutDetect"]
  debugExtractParams = hyperparameters["debugExtractParams"]
  tailAngleSmoothingFactor = hyperparameters["tailAngleSmoothingFactor"]

  trackingHeadTailAllAnimals = trackingData[0]
  trackingHeadingAllAnimals  = trackingData[1]
  trackingEyesAllAnimals     = trackingData[2]
  headPositionFirstFrame     = trackingData[3]
  tailTipFirstFrame          = trackingData[4]
  auDessusPerAnimal          = trackingData[5] if len(trackingData) == 6 else 0
    
  data = []
  
  for animalId in range(0, len(trackingHeadTailAllAnimals)):
    
    trackingTail    = trackingHeadTailAllAnimals[animalId]
    trackingHeading = trackingHeadingAllAnimals[animalId]
    
    n = len(trackingTail[0])
    
    debug = 0

    nbFrames = len(trackingTail)
    nbPoints = len(trackingTail[0])

    tail_1    = np.zeros((nbFrames, 2))
    tail_2    = np.zeros((nbFrames, 2))
    tip       = np.zeros((nbFrames, 2))
    head      = np.zeros((nbFrames, 2))
    heading   = np.zeros((nbFrames, 1))
    angle     = np.zeros((nbFrames, 1))
    allAngles = np.zeros((nbFrames, n))
    tailX     = np.zeros((nbFrames, n))
    tailY     = np.zeros((nbFrames, n))
    
    if hyperparameters["headingCalculationMethod"] == "calculatedWithMedianTailTip":
      tip2 = np.zeros((nbFrames, 2))
      nbTailPoints = len(trackingTail[0]) - 1
      for i in range(0,nbFrames):
        tip2[i] = np.array([ trackingTail[i][nbTailPoints][0], trackingTail[i][nbTailPoints][1] ])
      medianTip = np.median(tip2,axis=0)
      
    curvatureBef = 0
    
    for i in range(0,nbFrames):
    
      if (hyperparameters["freqAlgoPosFollow"] != 0) and (i % hyperparameters["freqAlgoPosFollow"] == 0):
        print("Extract Param Begin: wellNumber:",wellNumber," ; frame:",i)
      
      tail_1[i]  = np.array([ trackingTail[i][1][0], trackingTail[i][1][1] ])
      tail_2[i]  = np.array([ trackingTail[i][2][0], trackingTail[i][2][1] ])
      
      nbTailPoints = len(trackingTail[0]) - 1
      tip[i]     = np.array([ trackingTail[i][nbTailPoints][0], trackingTail[i][nbTailPoints][1] ])
      
      head[i]    = np.array([ trackingTail[i][0][0], trackingTail[i][0][1] ])
      
      if hyperparameters["headingCalculationMethod"] == "calculatedWithFirstTailPt":
        heading[i] = calculateAngle(head[i], tail_1[i])
      if hyperparameters["headingCalculationMethod"] == "calculatedWithTwoFirstTailPt":
        heading[i] = calculateAngle(head[i], tail_2[i])
      elif hyperparameters["headingCalculationMethod"] == "calculatedWithMedianTailTip":
        heading[i] = calculateAngle(head[i], medianTip)
      elif hyperparameters["headingCalculationMethod"] == "simplyFromPreviousCalculations":
        heading[i] = trackingHeading[i]
        heading[i] = (heading[i] + math.pi) % (2*math.pi)
      elif hyperparameters["headingCalculationMethod"] == "fromPreviousCalculationsAndAdjustWithPreviousFrame":
        heading[i] = trackingHeading[i]
        heading[i] = (heading[i] + math.pi) % (2*math.pi)
        if i != 0 and distBetweenThetas(heading[i-1], ((heading[i] + math.pi) % (2*math.pi))) > distBetweenThetas(heading[i-1], heading[i]):
          heading[i] = (heading[i] + math.pi) % (2*math.pi)
      else: # calculatedWithHead : THIS IS THE DEFAULT
        if not(math.isnan(trackingHeading[i])):
          heading[i] = trackingHeading[i]
          diff1 = distBetweenThetas(heading[i],           calculateAngle(head[i], tail_1[i]))
          diff2 = distBetweenThetas(heading[i] + math.pi, calculateAngle(head[i], tail_1[i]))
          if diff2 < diff1:
            heading[i] = heading[i] + math.pi
        else:
          heading[i] = 0
      
      if type(tailAngle) == int:
        angle[i] = calculateTailAngle(calculateAngle(head[i], tip[i]), heading[i])
      else:
        angle[i] = tailAngle[i]
        
      if hyperparameters["calculateAllTailAngles"]:
        for j in range(0, n):
          allAngles[i][j] = calculateTailAngle(calculateAngle(head[i], np.array([trackingTail[i][j][0], trackingTail[i][j][1]])), heading[i])
      
      heading[i] = (heading[i] + math.pi) % (2*math.pi)
      
      tailX[i]   = trackingTail[i,:,0]
      tailY[i]   = trackingTail[i,:,1]
    
    if hyperparameters["saveAllDataEvenIfNotInBouts"] or hyperparameters["storeH5"]:
      trackingFlatten = [trackingHeadTailAllAnimals[animalId][i].flatten().tolist() + [_calculateTailLength(tailX[i], tailY[i])] + heading[i].tolist() + angle[i].tolist() for i in range(0, len(heading))]
      trackingFlattenColumnsNames = ['HeadPosX', 'HeadPosY']
      for i in range(0, (len(trackingHeadTailAllAnimals[0][0].flatten().tolist()) - 2) // 2):
        trackingFlattenColumnsNames += ['TailPosX' + str(i + 1)]
        trackingFlattenColumnsNames += ['TailPosY' + str(i + 1)]
      trackingFlattenColumnsNames.append('TailLength')
      trackingFlattenColumnsNames += ['Heading']
      trackingFlattenColumnsNames += ['tailAngle']
    
    if hyperparameters["noBoutsDetection"] == 1:
      auDessus        = np.zeros((nbFrames, 1))
      for i in range(0,nbFrames):
        auDessus[i] = 1
    elif hyperparameters["boutEdgesWhereZeros"] == 1:
      auDessus        = np.zeros((nbFrames, 1))
      for i in range(0,nbFrames):
        if ((head[i][0] == 0) and (head[i][1] == 0)):
          auDessus[i] = 0
        else:
          auDessus[i] = 1
    elif type(auDessusPerAnimal) != int:

      auDessus = auDessusPerAnimal[animalId]
    elif hyperparameters["coordinatesOnlyBoutDetection"]:
      frameDistance = hyperparameters["frameGapComparision"]
      auDessus = [int(math.sqrt(sum((head[i+frameDistance] - coords) ** 2)) >= hyperparameters["coordinatesOnlyBoutDetectionMinDist"]) for i, coords in enumerate(head[:-frameDistance])]
      auDessus.extend([0] * frameDistance)
    elif hyperparameters["thresForDetectMovementWithRawVideo"] == 0:

      # Calculating angle median
      angle2 = np.transpose(angle)
      angle2 = angle2[0]
      rolling_window = hyperparameters["tailAngleMedianFilter"]
      if rolling_window > 0:
        shift = int(-rolling_window / 2)
        angle_median = np.array(pd.Series(angle2).rolling(rolling_window).median())
        angle_median = np.roll(angle_median, shift)
        for ii in range(0, rolling_window):
          angle_median[ii] = angle2[ii]
        for ii in range(len(angle_median)-rolling_window,len(angle_median)):
          angle_median[ii] = angle2[ii]
      else:
        angle_median = angle2
      # Calculating angle variation to detect movement
      angleVariation  = np.zeros((nbFrames, 1))
      auDessus        = np.zeros((nbFrames, 1))
      for i in range(0,nbFrames):
        if (hyperparameters["freqAlgoPosFollow"] != 0) and (i % (hyperparameters["freqAlgoPosFollow"]) == 0):
          print("Extract Param Middle: wellNumber:",wellNumber," ; frame:",i)
        min = 10000
        max = -10000
        # for j in range(i-hyperparameters["windowForBoutDetectWithAngle"], i+1):
        for j in range(i-int(hyperparameters["windowForBoutDetectWithAngle"]/2), i+int(hyperparameters["windowForBoutDetectWithAngle"]/2)):
          if (j >= 0) and (j<nbFrames):
            if angle_median[j] > max:
              max = angle_median[j]
            if angle_median[j] < min:
              min = angle_median[j]
        angleVariation[i] = max - min
        if debug:
          print(i,angleVariation[i])
        if angleVariation[i] > thresAngleBoutDetect:
          auDessus[i] = 1

    else:
      auDessus = detectMovementWithRawVideo(hyperparameters, videoPath, background, wellNumber, wellPositions, head, headPositionFirstFrame, tailTipFirstFrame)
    
    if (hyperparameters["freqAlgoPosFollow"] != 0):
      print("Extract Param Middle: wellNumber:",wellNumber," ; frame:",i)
    
    auDessus2 = np.copy(auDessus)
    if hyperparameters["boutEdgesWhereZeros"] == 0 and hyperparameters["noBoutsDetection"] == 0:
      windowGap = hyperparameters["fillGapFrameNb"]
      if windowGap:
        for i in range(windowGap,nbFrames-windowGap):
          if auDessus[i] == 0:
            j = i - windowGap
            while (auDessus[j] == 0) and (j < i + windowGap):
              j = j + 1
            if j < i + windowGap:
              j = j + 1;
              while (auDessus[j] == 0) and (j < i + windowGap):
                j = j + 1;
              if auDessus[j] > 0:
                for k in range(i - windowGap, i + windowGap):
                  auDessus2[k] = 1
    
    bouts   = np.zeros((0, 3))
    curBout = np.zeros((1, 3))
    position = 0
    
    for i in range(0, nbFrames):
    
      if (hyperparameters["freqAlgoPosFollow"] != 0) and (i % (hyperparameters["freqAlgoPosFollow"]*10) == 0):
        print("Extract Param End1 Freq*10 : wellNumber:",wellNumber," ; frame:",i)
    
      if (position==0) and (auDessus2[i]==1):
        curBout[0][0] = wellNumber
        curBout[0][1] = i
        position = 1
        retire = 1
      if (position == 1) and ((auDessus2[i] == 0) or (i == nbFrames-1)):
        curBout[0][2] = i-1
        position = 0
        if curBout[0][2] - curBout[0][1] >= hyperparameters["detectBoutMinNbFrames"] or (hyperparameters["noChecksForBoutSelectionInExtractParams"] and curBout[0][2] - curBout[0][1] >= hyperparameters["boutsMinNbFrames"]):
          debMouv = int(curBout[0][1])
          endMouv = int(curBout[0][2])
          dist = math.sqrt( (head[debMouv,0]-head[endMouv,0])**2 + (head[debMouv,1]-head[endMouv,1])**2 )
          if ((dist >= hyperparameters["detectBoutMinDist"]) and (np.max(angle[debMouv:endMouv])-np.min(angle[debMouv:endMouv]) >= hyperparameters["detectBoutMinAngleDiff"])) or hyperparameters["noChecksForBoutSelectionInExtractParams"]:
            bouts = np.append(bouts, curBout, axis=0)
          else:
            if debugExtractParams:
              print("Bout starting at", debMouv, " has a dist or angle diff too big. Dist:", dist," minimum was:", hyperparameters["detectBoutMinDist"],". DetectBoutMinAngleDiff:",np.max(angle[debMouv:endMouv])-np.min(angle[debMouv:endMouv]),", minimmum was:",hyperparameters["detectBoutMinAngleDiff"])
        else:
          if debugExtractParams:
            print("Bout starting at", int(curBout[0][1]), "was too small. Length:", curBout[0][2] - curBout[0][1],". Minimum was:", hyperparameters["detectBoutMinNbFrames"])
    
    # Refining beginning and end of bout detection to remove the artefacts from the fill gap procedure
    for numBout in range(0,len(bouts)):
      debMouv = int(bouts[numBout][1])
      finMouv = int(bouts[numBout][2])
      # Refining the beginning
      pos = debMouv
      while int(auDessus[pos]) == 0:
        pos = pos + 1
      bouts[numBout][1] = pos
      # Refining the end
      pos = finMouv
      while int(auDessus[pos]) == 0:
        pos = pos - 1
      bouts[numBout][2] = pos    
      if hyperparameters["addOneFrameAtTheEndForBoutDetection"]:
        if bouts[numBout][2] + 1 < nbFrames:
          bouts[numBout][2] = bouts[numBout][2] + 1

    if hyperparameters["saveAllDataEvenIfNotInBouts"] and not hyperparameters['noBoutsDetection']:
      for frameData in trackingFlatten:
        frameData.append(float('nan'))
      trackingFlattenColumnsNames.append('BoutNumber')

    for i in range(0,len(bouts)):
    
      if (hyperparameters["freqAlgoPosFollow"] != 0) and (i % (hyperparameters["freqAlgoPosFollow"]*10) == 0):
        print("Extract Param End2 Freq*10 : wellNumber:",wellNumber," ; frame:",i)
      
      item = {}
      start = int(bouts[i][1])
      end   = int(bouts[i][2])
      item["AnimalNumber"]  = animalId
      item["BoutStart"]     = start + firstFrame
      item["BoutEnd"]       = end + firstFrame
      item["TailAngle_Raw"] = angle[start:end+1,0].tolist()

      if hyperparameters["saveAllDataEvenIfNotInBouts"] and not hyperparameters['noBoutsDetection']:
        for frameData in trackingFlatten[start:end+1]:
            frameData[-1] = i

      if hyperparameters["eyeTracking"]:
        item["leftEyeX"]      = trackingEyesAllAnimals[animalId, start:end+1, 0].tolist()
        item["leftEyeY"]      = trackingEyesAllAnimals[animalId, start:end+1, 1].tolist()
        item["leftEyeAngle"]  = trackingEyesAllAnimals[animalId, start:end+1, 2].tolist()
        item["leftEyeArea"]   = trackingEyesAllAnimals[animalId, start:end+1, 3].tolist()
        item["rightEyeX"]     = trackingEyesAllAnimals[animalId, start:end+1, 4].tolist()
        item["rightEyeY"]     = trackingEyesAllAnimals[animalId, start:end+1, 5].tolist()
        item["rightEyeAngle"] = trackingEyesAllAnimals[animalId, start:end+1, 6].tolist()
        item["rightEyeArea"]  = trackingEyesAllAnimals[animalId, start:end+1, 7].tolist()
      
      if hyperparameters["calculateAllTailAngles"]:
        [tailangles_arr, tailangles_arr_smoothed] = smoothAllTailAngles(allAngles, hyperparameters, start, end)
        item["allTailAngles"]         = tailangles_arr.tolist()
        item["allTailAnglesSmoothed"] = tailangles_arr_smoothed.tolist()
      
      if np.isnan(head[start:end+1,0]).any():
        item["HeadX"] = [0]
        item["HeadY"] = [0]
      else:
        item["HeadX"]         = head[start:end+1,0].tolist()
        item["HeadY"]         = head[start:end+1,1].tolist()
        item["Heading"]       = heading[start:end+1,0].tolist()
        item["TailX_VideoReferential"] = tailX[start:end+1].tolist()
        item["TailY_VideoReferential"] = tailY[start:end+1].tolist()
      data.append(item)

    if hyperparameters["saveAllDataEvenIfNotInBouts"] or hyperparameters["storeH5"]:
      trackingFlattenPandas = pd.DataFrame(trackingFlatten, columns=trackingFlattenColumnsNames)
    if hyperparameters["saveAllDataEvenIfNotInBouts"]:
      outputFolder = os.path.join(hyperparameters["outputFolder"], hyperparameters["videoNameWithTimestamp"])
      if not os.path.exists(outputFolder):
        os.makedirs(outputFolder)
      fname = os.path.join(outputFolder, f'allData_{hyperparameters["videoName"]}_wellNumber{wellNumber}_animal{animalId}.csv')
      with open(fname, 'w+', newline='') as f:
        if hyperparameters["videoFPS"]:
          f.write(f'videoFPS: {hyperparameters["videoFPS"]}\n')
        if hyperparameters["videoPixelSize"]:
          f.write(f'videoPixelSize: {hyperparameters["videoPixelSize"]}\n')
        trackingFlattenPandas.convert_dtypes().to_csv(f)
    if hyperparameters['storeH5']:
      with h5py.File(hyperparameters['H5filename'], 'a') as results:
        group = results.create_group(f"dataForWell{wellNumber}/dataForAnimal{animalId}/dataPerFrame")
        datasets = {'HeadPos': (('HeadPosX', 'HeadPosY'), ('X', 'Y')),
                    'Heading': 'Heading',
                    'TailAngle': 'tailAngle',
                    'TailPosX': (tuple(f'TailPosX{pos}' for pos in range(1, nbPoints)), tuple(f'Pos{pos}' for pos in range(1, nbPoints))),
                    'TailPosY': (tuple(f'TailPosY{pos}' for pos in range(1, nbPoints)), tuple(f'Pos{pos}' for pos in range(1, nbPoints))),
                    'TailLength': 'TailLength'}
        for name, columns in datasets.items():
          if isinstance(columns, str):  # 1d array
            group.create_dataset(name, data=trackingFlattenPandas[columns].values)
          else:
            assert isinstance(columns, tuple)  # 2d array
            pandasNames, h5Names = columns
            headPosData = np.empty(len(trackingFlattenPandas.index), dtype=list(zip(h5Names, (trackingFlattenPandas[name].dtype for name in pandasNames))))
            for h5Name, pandasName in zip(h5Names, pandasNames):
              headPosData[h5Name] = trackingFlattenPandas[pandasName].values
            headPos = group.create_dataset(name, data=headPosData)
            headPos.attrs['columns'] = headPosData.dtype.names

  print("Parameters extracted for well",wellNumber)
  if hyperparameters["popUpAlgoFollow"]:
    import zebrazoom.code.popUpAlgoFollow as popUpAlgoFollow
    popUpAlgoFollow.prepend("Parameters extracted for well " + str(wellNumber))
  
  return data



import h5py
import cv2
import os
import shutil
import math
import matplotlib.pyplot as plt
import numpy as np
import time
import pandas as pd
import pickle
import scipy as sp

def tailAnglesHeatMap(superStruct, hyperparameters, videoNameWithTimestamp):
  
  # Creation of the sub-folder "anglesHeatMap" 
  outputPath = os.path.join(hyperparameters["outputFolder"], videoNameWithTimestamp, "anglesHeatMap")
  if not os.path.exists(outputPath):
    os.makedirs(outputPath)
  
  # Video frame start and end
  firstFrame = hyperparameters["firstFrame"]
  lastFrame  = hyperparameters["lastFrame"]
  nbFrames   = lastFrame - firstFrame + 1
  perBoutOutputVideoStartStopFrameMargin = hyperparameters["perBoutOutputVideoStartStopFrameMargin"]
  
  plotTailAngleRaw      = False
  plotTailAngleSmoothed = False
  plotTailAngleHeatmap  = True
  
  pointsToTakeIntoAccountStart = 9 - int(hyperparameters["tailAnglesHeatMapNbPointsToTakeIntoAccount"])
  
  # Going through each well, each fish and each bout
  for i in range(0, len(superStruct["wellPoissMouv"])):
    for j in range(0, len(superStruct["wellPoissMouv"][i])):
      tailAngleHeatmaps = {}
      if hyperparameters["saveAllDataEvenIfNotInBouts"]:
        fname = os.path.join(hyperparameters["outputFolder"], videoNameWithTimestamp, f'allData_{hyperparameters["videoName"]}_wellNumber{i}_animal{j}.csv')
        startLines = []
        with open(fname) as f:
          line = f.readline()
          while ',' not in line:
            startLines.append(line)
            line = f.readline()
        df = pd.read_csv(fname, skiprows=len(startLines))
        df = df.loc[:, ~df.columns.str.contains('^Unnamed')]
      for k in range(0, len(superStruct["wellPoissMouv"][i][j])):
        
        # Creation of tail angles raw graph for bout k
        if plotTailAngleRaw:
          bStart = superStruct["wellPoissMouv"][i][j][k]["BoutStart"]
          bEnd   = superStruct["wellPoissMouv"][i][j][k]["BoutEnd"]
          tailAngles = superStruct["wellPoissMouv"][i][j][k]["allTailAngles"][pointsToTakeIntoAccountStart:]
          plt.figure(1)
          for tailAngle in tailAngles:
            if bEnd - bStart + 1 == len(tailAngle):
              plt.plot([i for i in range(bStart, bEnd + 1)], [t*(180/math.pi) for t in tailAngle])
              if hyperparameters["perBoutOutputYaxis"]:
                plt.axis([bStart, bEnd + 1, hyperparameters["perBoutOutputYaxis"][0], hyperparameters["perBoutOutputYaxis"][1]])
          plt.savefig(os.path.join(outputPath, hyperparameters["videoName"] + "_angles_bout" + str(i) + '_' + str(j) + '_' + str(k) + '.png'))
          plt.close(1)
        
        # Creation of tail angles smoothed graph and tail angle heatmap for bout k
        bStart = superStruct["wellPoissMouv"][i][j][k]["BoutStart"]
        bEnd   = superStruct["wellPoissMouv"][i][j][k]["BoutEnd"]
        tailAngles = superStruct["wellPoissMouv"][i][j][k]["allTailAnglesSmoothed"][pointsToTakeIntoAccountStart:]
        tailAngleHeatmap = []
        if plotTailAngleSmoothed:
          plt.figure(1)
        for tailAngle in tailAngles:
          if bEnd - bStart + 1 == len(tailAngle):
            tailAngleHeatmap.append([t*(180/math.pi) for t in tailAngle])
            if plotTailAngleSmoothed:
              plt.plot([i for i in range(bStart, bEnd + 1)], [t*(180/math.pi) for t in tailAngle])
              if hyperparameters["perBoutOutputYaxis"]:
                plt.axis([bStart, bEnd + 1, hyperparameters["perBoutOutputYaxis"][0], hyperparameters["perBoutOutputYaxis"][1]])
        if plotTailAngleSmoothed:
          plt.savefig(os.path.join(outputPath, hyperparameters["videoName"] + "_anglesSmoothed_bout" + str(i) + '_' + str(j) + '_' + str(k) + '.png'))
          plt.close(1)

        if hyperparameters["saveAllDataEvenIfNotInBouts"] or hyperparameters["storeH5"]:
          tailAngleHeatmaps[(bStart - firstFrame, bEnd + 1 - firstFrame)] = tailAngleHeatmap

        # Creation of tail angle heatmap
        if plotTailAngleHeatmap:
          fig = plt.figure(1)
          maxAngle = np.max(np.abs(tailAngleHeatmap))
          # plt.pcolor(tailAngleHeatmap)
          plt.pcolor(tailAngleHeatmap[::-1], vmin=-maxAngle, vmax=maxAngle)
          ax = fig.axes
          ax[0].set_xlabel('Frame number')
          ax[0].set_ylabel('Tail angle: Tail base to tail extremity')
          plt.colorbar()
          plt.savefig(os.path.join(outputPath, hyperparameters["videoName"] + "_tailAngleHeatmap_bout" + str(i) + '_' + str(j) + '_' + str(k) + '.png'))
          plt.close(1)

      if hyperparameters["saveAllDataEvenIfNotInBouts"] or hyperparameters["storeH5"]:
        angleCount = max(map(len, tailAngleHeatmaps.values())) if tailAngleHeatmaps else 0
        tailAngleHeatmapData = [[float('nan')] * nbFrames for _ in range(angleCount)]
        for (start, end), values in tailAngleHeatmaps.items():
          for data, vals in zip(tailAngleHeatmapData, values):
            data[start:end] = vals
      if hyperparameters["saveAllDataEvenIfNotInBouts"]:
        for idx, data in enumerate(tailAngleHeatmapData):
          df[f'tailAngleHeatmap{idx + 1}'] = data
        with open(fname, 'w+', newline='') as f:
          f.write(''.join(startLines))
          df.convert_dtypes().to_csv(f)
      if hyperparameters['storeH5'] and angleCount:
        with h5py.File(hyperparameters['H5filename'], 'a') as results:
          arr = np.empty(nbFrames, dtype=[(f'Pos{idx}', float) for idx in range(1, angleCount + 1)])
          for idx, data in enumerate(tailAngleHeatmapData):
            arr[f'Pos{idx + 1}'] = data
          dataset = results.create_dataset(f"dataForWell{i}/dataForAnimal{j}/dataPerFrame/tailAngleHeatmap", data=arr)
          dataset.attrs['columns'] = arr.dtype.names



import numpy as np
import queue
import math
import cv2

def detectMovementWithRawVideoInsideTracking(self, i, grey):
  if self._previousFrames is None:
    self._previousFrames = queue.Queue(self._hyperparameters["frameGapComparision"])
    self._auDessusPerAnimalIdList = [[np.zeros((self._lastFrame-self._firstFrame+1, 1)) for nbAnimalsPerWell in range(0, self._hyperparameters["nbAnimalsPerWell"])]
                                     for wellNumber in range(len(self._wellPositions))]
  halfDiameterRoiBoutDetect = self._hyperparameters["halfDiameterRoiBoutDetect"]
  if self._previousFrames.full():
    previousFrame   = self._previousFrames.get()
    curFrame        = grey
    for wellNumber in self._listOfWellsOnWhichToRunTheTracking:
      for animal_Id in range(self._hyperparameters["nbAnimalsPerWell"]):
        headX = self._trackingDataPerWell[wellNumber][animal_Id][i-1][0][0]
        headY = self._trackingDataPerWell[wellNumber][animal_Id][i-1][0][1]
        xmin = headX - self._hyperparameters["halfDiameterRoiBoutDetect"]
        ymin = headY - self._hyperparameters["halfDiameterRoiBoutDetect"]
        xmax = xmin + 2 * self._hyperparameters["halfDiameterRoiBoutDetect"]
        ymax = ymin + 2 * self._hyperparameters["halfDiameterRoiBoutDetect"]
        lenX = self._wellPositions[wellNumber]['lengthX']
        lenY = self._wellPositions[wellNumber]['lengthY']
        if xmin < 0:
          xmin = 0
        if ymin < 0:
          ymin = 0
        if xmax > lenX - 1:
          xmax = lenX - 1
        if ymax > lenY - 1:
          ymax = lenY - 1
        if ymax < ymin:
          ymax = ymin + 2 * self._hyperparameters["halfDiameterRoiBoutDetect"]
        if xmax < xmin:
          xmax = xmin + 2 * self._hyperparameters["halfDiameterRoiBoutDetect"]
        if ( (xmin > lenX - 1) or (xmax < 0) ):
          xmin = 0
          xmax = 0 + lenX - 1
        if ( (ymin > lenY - 1) or (ymax < 0) ):
          ymin = 0
          ymax = 0 + lenY - 1
        xmin = int(xmin + self._wellPositions[wellNumber]['topLeftX'])
        xmax = int(xmax + self._wellPositions[wellNumber]['topLeftX'])
        ymin = int(ymin + self._wellPositions[wellNumber]['topLeftY'])
        ymax = int(ymax + self._wellPositions[wellNumber]['topLeftY'])
        subPreviousFrame = previousFrame[ymin:ymax, xmin:xmax].copy()
        subCurFrame      = curFrame[ymin:ymax, xmin:xmax].copy()

        res = cv2.absdiff(subPreviousFrame, subCurFrame)
        
        ret, res = cv2.threshold(res,self._hyperparameters["thresForDetectMovementWithRawVideo"],255,cv2.THRESH_BINARY)
        totDiff = cv2.countNonZero(res)

        if totDiff > self._hyperparameters["minNbPixelForDetectMovementWithRawVideo"]:
          self._auDessusPerAnimalIdList[wellNumber][animal_Id][i] = 1
        else:
          self._auDessusPerAnimalIdList[wellNumber][animal_Id][i] = 0
  
  else:
    
    for wellNumber in self._listOfWellsOnWhichToRunTheTracking:
      for animalId in range(0, self._hyperparameters["nbAnimalsPerWell"]):
        self._auDessusPerAnimalIdList[wellNumber][animalId][i] = 0 # Need to improve this part
  
  self._previousFrames.put(grey)



import zebrazoom
import zebrazoom.code.tracking
import zebrazoom.code.tracking.customTrackingImplementations
from zebrazoom.code.findWells import findWells
from zebrazoom.code.createSuperStruct import createSuperStruct
from zebrazoom.code.createValidationVideo import createValidationVideo
from zebrazoom.code.getHyperparameters import getHyperparameters

import h5py
import pickle
import os
import shutil
import time
import cv2
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
import json
import glob
import numpy as np
from datetime import datetime


class ZebraZoomVideoAnalysis:
  def __init__(self, pathToVideo, videoName, videoExt, configFile, argv, useGUI=True):
    self._pathToVideo = pathToVideo
    self._videoName = videoName
    self._configFile = configFile
    self._useGUI = useGUI
    self._videoNameWithExt = videoName + '.' + videoExt
    self.wellPositions = None
    # Getting hyperparameters
    self._hyperparameters, self._configFile = getHyperparameters(configFile, self._videoNameWithExt, os.path.join(pathToVideo, self._videoNameWithExt), argv)
    videoNameWithTimestamp = f'{self._hyperparameters["videoName"]}_{datetime.now().strftime("%Y_%m_%d-%H_%M_%S")}'
    while os.path.exists(os.path.join(self._hyperparameters["outputFolder"], f'{videoNameWithTimestamp}.h5')) or os.path.exists(os.path.join(self._hyperparameters["outputFolder"], f'{videoNameWithTimestamp}')):
      time.sleep(0.5)
      videoNameWithTimestamp = f'{self._hyperparameters["videoName"]}_{datetime.now().strftime("%Y_%m_%d-%H_%M_%S")}'
    self._hyperparameters['videoNameWithTimestamp'] = videoNameWithTimestamp
    if self._hyperparameters['storeH5']:
      self._hyperparameters['H5filename'] = os.path.join(self._hyperparameters["outputFolder"], f'{videoNameWithTimestamp}.h5')
    # Setting output folder
    self._outputFolderVideo = os.path.join(self._hyperparameters["outputFolder"], videoNameWithTimestamp)

  def _checkFirstAndLastFrame(self):
    # Checking first frame and last frame value
    cap   = zzVideoReading.VideoCapture(os.path.join(self._pathToVideo, self._videoNameWithExt))
    nbFrames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    cap.release()
    if self._hyperparameters["firstFrame"] < 0:
      print("Error for video " + self._videoName + ": The parameter 'firstFrame' in your configuration file is too small" + " (firstFrame value is " + str(self._hyperparameters["firstFrame"]) + ", number of frames in the video is " + str(nbFrames) + ")")
      raise NameError("Error for video " + self._videoName + ": The parameter 'firstFrame' in your configuration file is too small" + " (firstFrame value is " + str(self._hyperparameters["firstFrame"]) + ", number of frames in the video is " + str(nbFrames) + ")")
    if self._hyperparameters["firstFrame"] > nbFrames:
      print("Error for video " + self._videoName + ": The parameter 'firstFrame' in your configuration file is too big" + " (firstFrame value is " + str(self._hyperparameters["firstFrame"]) + ", number of frames in the video is " + str(nbFrames) + ")")
      raise NameError("Error for video " + self._videoName + ": The parameter 'firstFrame' in your configuration file is too big" + " (firstFrame value is " + str(self._hyperparameters["firstFrame"]) + ", number of frames in the video is " + str(nbFrames) + ")")
    if (self._hyperparameters["lastFrame"] < 0) or (self._hyperparameters["lastFrame"] <= self._hyperparameters["firstFrame"]):
      print("Error for video " + self._videoName + ": The parameter 'lastFrame' in your configuration file is too small" + " (lastFrame value is " + str(self._hyperparameters["lastFrame"]) + ", number of frames in the video is " + str(nbFrames) + ")")
      raise NameError("Error for video " + self._videoName + ": The parameter 'lastFrame' in your configuration file is too small" + " (lastFrame value is " + str(self._hyperparameters["lastFrame"]) + ", number of frames in the video is " + str(nbFrames) + ")")
    if self._hyperparameters["lastFrame"] > nbFrames:
      print("Warning: The parameter 'lastFrame' in your configuration file is too big so we adjusted it to the value:", nbFrames-2, "it was originally set to", self._hyperparameters["lastFrame"])
      self._hyperparameters["lastFrame"] = nbFrames - 2
      # raise NameError("Error: The parameter 'lastFrame' in your configuration file is too big")

  def _storeConfigUsed(self):
    '''Saving the configuration file used'''
    if self._hyperparameters['storeH5']:
      with h5py.File(self._hyperparameters['H5filename'], 'a') as results:
        results.require_group("configurationFileUsed").attrs.update(self._configFile)
    else:
      if not os.path.exists(self._outputFolderVideo):
        os.makedirs(self._outputFolderVideo)
      with open(os.path.join(self._outputFolderVideo, 'configUsed.json'), 'w') as outfile:
        json.dump(self._configFile, outfile)

  def getWellPositions(self):
    '''Get well positions'''
    rotationAngleParams = None
    if self._hyperparameters["headEmbeded"] and not self._hyperparameters["oneWellManuallyChosenTopLeft"]:
      self.wellPositions = [{"topLeftX":0, "topLeftY":0, "lengthX": self._hyperparameters["videoWidth"], "lengthY": self._hyperparameters["videoHeight"]}]
    else:
      print("start find wells")
      maybeReloadWells = self._hyperparameters["groupOfMultipleSameSizeAndShapeEquallySpacedWells"] or self._hyperparameters["multipleROIsDefinedDuringExecution"]
      inputFilesFolder = os.path.join(self._hyperparameters['outputFolder'], '.ZebraZoomVideoInputs', self._videoName)
      if self._hyperparameters["groupOfMultipleSameSizeAndShapeEquallySpacedWells"]:
        rotationAngleFile = os.path.join(inputFilesFolder, 'rotationAngle.txt')
        if os.path.exists(rotationAngleFile):
          with open(rotationAngleFile, 'rb') as f:
            rotationAngleParams = pickle.load(f)
          self._hyperparameters.update(rotationAngleParams)
      if self._hyperparameters["saveWellPositionsToBeReloadedNoMatterWhat"]:
        if not maybeReloadWells:
          print('Well positions can only be stored for multiple ROIs defined during execution or grid system')
          raise ValueError
        self.wellPositions = findWells(os.path.join(self._pathToVideo, self._videoNameWithExt), self._hyperparameters)
        if not os.path.exists(inputFilesFolder):
          os.makedirs(inputFilesFolder)
        with open(os.path.join(inputFilesFolder, 'intermediaryWellPositionReloadNoMatterWhat.txt'), 'wb') as outfile:
          pickle.dump(self.wellPositions, outfile)
      elif maybeReloadWells and os.path.exists(os.path.join(inputFilesFolder, 'intermediaryWellPositionReloadNoMatterWhat.txt')):
        with open(os.path.join(inputFilesFolder, 'intermediaryWellPositionReloadNoMatterWhat.txt'), 'rb') as outfile:
          self.wellPositions = pickle.load(outfile)
        gridSystem = len(self.wellPositions) > 1 and len({(pos['lengthX'], pos['lengthY']) for pos in self.wellPositions}) == 1
        nbWells = self._hyperparameters["nbWellsPerRows"] * self._hyperparameters["nbRowsOfWells"] if self._hyperparameters["groupOfMultipleSameSizeAndShapeEquallySpacedWells"] else self._hyperparameters["nbWells"]
        if self._hyperparameters["groupOfMultipleSameSizeAndShapeEquallySpacedWells"] ^ gridSystem or len(self.wellPositions) != nbWells:
          self.wellPositions = findWells(os.path.join(self._pathToVideo, self._videoNameWithExt), self._hyperparameters)
      elif self._hyperparameters["reloadWellPositions"]:
        fname = next(reversed(sorted(name for name in os.listdir(self._hyperparameters['outputFolder']) if os.path.splitext(name)[0][:-20] == self._videoName and os.path.splitext(name)[0] != self._hyperparameters['videoNameWithTimestamp'])))
        with h5py.File(os.path.join(self._hyperparameters['outputFolder'], fname)) as results:
          self.wellPositions = [dict(results[f'wellPositions/well{idx}'].attrs) for idx in range(len(results['wellPositions']))]
      elif self._hyperparameters["reloadWellPositionsFromFileInZZoutputIfItExistSaveInItOtherwise"] and os.path.exists(os.path.join(self._hyperparameters["outputFolder"], 'wellPosition.txt')):
        with open(os.path.join(self._hyperparameters["outputFolder"], 'wellPosition.txt'), 'rb') as outfile:
          self.wellPositions = pickle.load(outfile)
      else:
        self.wellPositions = findWells(os.path.join(self._pathToVideo, self._videoNameWithExt), self._hyperparameters)
        if self._hyperparameters["reloadWellPositionsFromFileInZZoutputIfItExistSaveInItOtherwise"]:
          with open(os.path.join(self._hyperparameters["outputFolder"], 'wellPosition.txt'), 'wb') as outfile:
            pickle.dump(self.wellPositions, outfile)

      if self._useGUI:
        from PyQt5.QtWidgets import QApplication

        app = QApplication.instance()
        if hasattr(app, "wellPositions"):
          if self.wellPositions is None:
            return
          app.wellPositions[:] = [(position['topLeftX'], position['topLeftY'], position['lengthX'], position['lengthY'])
                                  for idx, position in enumerate(self.wellPositions)]
          if self._hyperparameters["wellsAreRectangles"] or len(self._hyperparameters["oneWellManuallyChosenTopLeft"]) or int(self._hyperparameters["multipleROIsDefinedDuringExecution"]) or self._hyperparameters["noWellDetection"] or self._hyperparameters["groupOfMultipleSameSizeAndShapeEquallySpacedWells"]:
            app.wellShape = 'rectangle'
          else:
            app.wellShape = 'circle'

    if self.wellPositions is not None:
      if self._hyperparameters['storeH5']:
        with h5py.File(self._hyperparameters['H5filename'], 'a') as results:
          for idx, wellPositions in enumerate(self.wellPositions):
            results.require_group(f"wellPositions/well{idx}").attrs.update(wellPositions)
          if rotationAngleParams is not None:
            results['configurationFileUsed'].attrs.update(rotationAngleParams)
      else:
        with open(os.path.join(self._outputFolderVideo, 'intermediaryWellPosition.txt'), 'wb') as outfile:
          pickle.dump(self.wellPositions, outfile)
        if rotationAngleParams is not None:
          with open(os.path.join(self._outputFolderVideo, 'configUsed.json'), 'r') as f:
            config = json.load(f)
          config.update(rotationAngleParams)
          with open(os.path.join(self._outputFolderVideo, 'configUsed.json'), 'w') as f:
            json.dump(config, f)

  def _runTracking(self):
    if 'trackingImplementation' in self._hyperparameters:
      name = self._hyperparameters['trackingImplementation']
    else:
      name = "fasterMultiprocessing" if self._hyperparameters["fasterMultiprocessing"] == 1 else "fasterMultiprocessing2" if self._hyperparameters["fasterMultiprocessing"] == 2 else "tracking"
    tracking = zebrazoom.code.tracking.get_tracking_method(name)(os.path.join(self._pathToVideo, self._videoNameWithExt), self.wellPositions, self._hyperparameters)
    if hasattr(tracking, 'useGUI'):
      tracking.useGUI = self._useGUI
    return tracking.run(), getattr(tracking, 'dataPostProcessing', None)

  def _createSuperStruct(self, paramDataPerWell):
    '''Create super structure'''
    return createSuperStruct(paramDataPerWell, self.wellPositions, self._hyperparameters, os.path.join(self._pathToVideo, self._videoNameWithExt))

  def _createValidationVideo(self, superStruct):
    '''Create validation video'''
    if self._hyperparameters.get('savePathToOriginalVideoForValidationVideo', False):
      return
    if self._hyperparameters["copyOriginalVideoToOutputFolderForValidation"]:
      fname = f'{self._hyperparameters["videoNameWithTimestamp"]}_originalVideoWithoutAnyTrackingDisplayed_pleaseUseTheGUIToVisualizeTrackingPoints.avi'
      shutil.copyfile(os.path.join(self._pathToVideo, self._videoNameWithExt), fname)
    elif self._hyperparameters["createValidationVideo"]:
      folder = self._hyperparameters['outputFolder'] if self._hyperparameters['storeH5'] else os.path.join(self._outputFolderVideo)
      if not os.path.exists(folder):
        os.makedirs(folder)
      fname = os.path.join(folder, f'{self._hyperparameters["videoNameWithTimestamp"]}.avi')
      infoFrame = createValidationVideo(os.path.join(self._pathToVideo, self._videoNameWithExt), superStruct, self._hyperparameters, outputName=fname)

  def _storeH5Results(self, superStruct):
    print("Store results:", self._hyperparameters['H5filename'])
    with h5py.File(self._hyperparameters['H5filename'], 'a') as results:
      results.attrs['version'] = 0
      results.attrs['firstFrame'] = superStruct["firstFrame"]
      results.attrs['lastFrame'] = superStruct['lastFrame']
      if 'videoFPS' in superStruct:
        results.attrs['videoFPS'] = superStruct['videoFPS']
      if 'videoPixelSize' in superStruct:
        results.attrs['videoPixelSize'] = superStruct['videoPixelSize']
      if 'pathToOriginalVideo' in superStruct:
        results.attrs['pathToOriginalVideo'] = superStruct['pathToOriginalVideo']
      keysToSkip = {'AnimalNumber', 'curvature', 'HeadX', 'HeadY', 'Heading', 'TailAngle_Raw', 'TailX_VideoReferential', 'TailY_VideoReferential'}
      for wellIdx, well in enumerate(superStruct['wellPoissMouv']):
        for animalIdx, animal in enumerate(well):
          listOfBouts = results.require_group(f"dataForWell{wellIdx}/dataForAnimal{animalIdx}/listOfBouts")
          listOfBouts.attrs['numberOfBouts'] = len(animal)
          for boutIdx, bout in enumerate(animal):
            boutGroup = listOfBouts.require_group(f'bout{boutIdx}')
            for key, value in bout.items():
              if key in keysToSkip:
                continue
              if isinstance(value, list):
                boutGroup.create_dataset(key, data=np.array(value))
              else:
                boutGroup.attrs[key] = value
      results.create_dataset('exampleFrame', data=zzVideoReading.VideoCapture(os.path.join(self._pathToVideo, self._videoNameWithExt)).read()[1])

  def _storeResults(self, superStruct):
    if self._hyperparameters['storeH5']:
      self._storeH5Results(superStruct)
    else:
      if not os.path.exists(self._outputFolderVideo):
        os.makedirs(self._outputFolderVideo)
      cv2.imwrite(os.path.join(self._outputFolderVideo, "exampleFrame.png"), zzVideoReading.VideoCapture(os.path.join(self._pathToVideo, self._videoNameWithExt)).read()[1])
      path = os.path.join(self._outputFolderVideo, f'results_{self._hyperparameters["videoNameWithTimestamp"]}.txt')
      print("createSuperStruct:", path)
      with open(path, 'w') as outfile:
        json.dump(superStruct, outfile)

    if self._hyperparameters["saveSuperStructToMatlab"]:
      from scipy.io import savemat
      if not os.path.exists(self._outputFolderVideo):
        os.makedirs(self._outputFolderVideo)
      matlabPath = os.path.join(self._outputFolderVideo, f'results_{self._hyperparameters["videoName"]}.mat')
      savemat(matlabPath, {'videoDataResults': superStruct})

  def _storeVersionUsed(self):
    if self._hyperparameters['storeH5']:
      with h5py.File(self._hyperparameters['H5filename'], 'a') as results:
        results.attrs['ZebraZoomVersionUsed'] = zebrazoom.__version__
    else:
      with open(os.path.join(self._outputFolderVideo, 'ZebraZoomVersionUsed.txt'), 'w') as fp:
        fp.write(zebrazoom.__version__)

  def _storeInAdditionalFolder(self):
      if os.path.isdir(self._hyperparameters["additionalOutputFolder"]):
        if self._hyperparameters["additionalOutputFolderOverwriteIfAlreadyExist"]:
          shutil.rmtree(self.hyperparameters["additionalOutputFolder"])
        else:
          print("The path " + self._hyperparameters["additionalOutputFolder"] + " already exists. New folder not created. If you want the folder to be overwritten in such situation in future executions, set the parameter 'additionalOutputFolderOverwriteIfAlreadyExist' to 1 in your configuration file.")
          return
      if os.path.exists(self._outputFolderVideo):
        shutil.copytree(self._outputFolderVideo, self._hyperparameters["additionalOutputFolder"])
      else:
        os.makedirs(self._hyperparameters["additionalOutputFolder"])
      if self._hyperparameters['storeH5']:
        for fname in os.listdir(self._hyperparameters['outputFolder']):
          fullPath = os.path.join(self._hyperparameters['outputFolder'], fname)
          if os.path.isfile(fullPath) and os.path.splitext(fname)[0] == self._hyperparameters['videoNameWithTimestamp']:
            shutil.copy2(fullPath, self._hyperparameters['additionalOutputFolder'])

  def run(self):
    '''Run tracking'''
    # Checking that path and video exists
    if not(os.path.exists(os.path.join(self._pathToVideo, self._videoNameWithExt))):
      print("Path or video name is incorrect for", os.path.join(self._pathToVideo, self._videoNameWithExt))
      return 0

    self._checkFirstAndLastFrame()

    self._storeConfigUsed()

    if self._hyperparameters["trackingDL"]:
      from torch.multiprocessing import Process
    else:
      from multiprocessing import Process

    # Launching GUI algoFollower if necessary
    if self._hyperparameters["popUpAlgoFollow"]:
      import zebrazoom.code.popUpAlgoFollow as popUpAlgoFollow

      popUpAlgoFollow.createTraceFile("starting ZebraZoom analysis on " + self._videoName)
      p = Process(target=popUpAlgoFollow.initialise)
      p.start()

    self.getWellPositions()
    if int(self._hyperparameters["exitAfterWellsDetection"]):
      if self._hyperparameters["saveWellPositionsToBeReloadedNoMatterWhat"]:
        try:  # try to clean up temporary results
          if self._hyperparameters['storeH5']:
            os.remove(self._hyperparameters['H5filename'])
          else:
            shutil.rmtree(os.path.join(self._hyperparameters['outputFolder'], self._hyperparameters['videoNameWithTimestamp']))
        except OSError:
          pass
      print("exitAfterWellsDetection")
      if self._hyperparameters["popUpAlgoFollow"]:
        import zebrazoom.code.popUpAlgoFollow as popUpAlgoFollow

        popUpAlgoFollow.prepend("ZebraZoom Analysis finished for " + self._videoName)
      raise ValueError

    paramDataPerWell, postProcessingCb = self._runTracking()

    superStruct = self._createSuperStruct(paramDataPerWell)
    self._createValidationVideo(superStruct)
    if postProcessingCb is not None:
      superStruct = postProcessingCb(self._outputFolderVideo, superStruct)
    self._storeResults(superStruct)

    self._storeVersionUsed()

    # Copying output result folder in another folder
    if len(self._hyperparameters["additionalOutputFolder"]):
      self._storeInAdditionalFolder()

    if self._hyperparameters["popUpAlgoFollow"]:
      import zebrazoom.code.popUpAlgoFollow as popUpAlgoFollow

      popUpAlgoFollow.prepend("ZebraZoom Analysis finished for " + self._videoName)
      # popUpAlgoFollow.prepend("")
      # if self._hyperparameters["closePopUpWindowAtTheEnd"]:
        # popUpAlgoFollow.prepend("ZebraZoom Analysis all finished")
